"""Pre-process input text, tokenizing, building vocabs, and pre-train word
level vectors."""

import os
import logging
from pickle import NONE
import warnings
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
import signal
import threading
from tqdm import tqdm
from functools import partial
from collections import namedtuple
from copy import copy
import json
import math
import queue
import random
from PIL import Image
from torchvision.transforms.functional import to_grayscale
from torchvision.transforms.functional import to_tensor
from gensim.models import Word2Vec
from ..SIF.segment.segment import FigureSegment
from ..SIF.segment import seg
from ..SIF.tokenization import tokenize
from ..SIF import EDU_SPYMBOLS
from ..ModelZoo.quesnet import QuesNetForPreTraining, AE
from .pretrian_utils import PretrainedEduTokenizer
from EduNLP import logger
import linecache
from typing import List, Union, Optional

Question = namedtuple('Question',
                      ['id', 'content', 'answer', 'false_options', 'labels'])


def save_list(item2index, path):
    item2index = sorted(item2index.items(), key=lambda kv: kv[1])
    items = [item for item, _ in item2index]
    with open(path, "wt", encoding="utf-8") as file:
        file.write('\n'.join(items))
    return


class QuesNetTokenizer(PretrainedEduTokenizer):
    """
    Examples
    --------
    >>> tokenizer = QuesNetTokenizer(meta=['knowledge'])
    >>> test_items = [{"ques_content": "$\\triangle A B C$ 的内角为 $A, \\quad B, $\\FigureID{test_id}$",
    ... "knowledge": "['*', '-', '/']"}, {"ques_content": "$\\triangle A B C$ 的内角为 $A, \\quad B",
    ... "knowledge": "['*', '-', '/']"}]
    >>> tokenizer.set_vocab(test_items,
    ... trim_min_count=1, key=lambda x: x["ques_content"], silent=True)
    >>> tokenizer.set_meta_vocab(test_items, silent=True)
    >>> token_items = [tokenizer(i, key=lambda x: x["ques_content"]) for i in test_items]
    >>> print(token_items[0].keys())
    dict_keys(['seq_idx', 'meta_idx'])
    >>> token_items = tokenizer(test_items, key=lambda x: x["ques_content"])
    >>> print(len(token_items["seq_idx"]))
    2
    """

    def __init__(self, vocab_path=None, meta_vocab_dir=None, img_dir: str = None,
                 max_length=250, tokenize_method="custom", symbol="mas", add_specials: list = None,
                 meta: List[str] = None, img_token='<img>', unk_token="<unk>", pad_token="<pad>", **kwargs):
        """
        Parameters
        ----------
        img_dir : str, optional
            path of images, by default None
        vocab_path : str, optional
            path of vacab file, by default None
        max_length : int, optional
            by default 250
        meta : list, optional
            the name of meta (side information), by default 'knowledge'
        img_token : str, optional
            by default '<img>'
        unk_token : str, optional
            by default "<unk>"
        pad_token : str, optional
            by default "<pad>"
        """
        if add_specials is None:
            add_specials = [img_token]
        else:
            add_specials = [img_token] + add_specials
        self.tokenization_params = {
            "formula_params": {
                "method": "linear",
                "skip_figure_formula": True
            }
        }
        kwargs.update(self.tokenization_params)
        super().__init__(vocab_path=vocab_path, max_length=max_length, tokenize_method=tokenize_method,
                         add_specials=add_specials, unk_token=unk_token, pad_token=pad_token, symbol=symbol, **kwargs)
        if meta is None:
            meta = ['know_name']
        self.img_dir = img_dir
        self.img_token = img_token
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.meta = meta
        self.meta_vocab_dir = meta_vocab_dir

        self.stoi = dict()
        self.itos = dict()
        self.stoi['word'] = self.vocab.token_to_idx
        self.itos['word'] = self.vocab.idx_to_token
        if meta_vocab_dir is not None:
            self.load_meta_vocab(meta_vocab_dir=meta_vocab_dir)
        self.config = {
            k: v for k, v in locals().items() if k not in [
                "self", "__class__", "vocab_path", 'img_dir', 'meta_vocab_dir']
        }

    def __call__(self, item: Union[str, dict, list], key=lambda x: x,
                 meta: Optional[list] = None,
                 padding=False, return_text=False, *args, **kwargs):
        """
        item: str or dict
            the question item
        key: function
            determine how to get the text of each item
        padding: bool
            whether to pad the seq_idx
        return_text: bool
            whether to return text tokens
        """
        if isinstance(item, list):
            ret = {
                "seq_idx": [],
                "meta_idx": []
            }
            if return_text:
                ret["seq_token"] = []
                ret["meta"] = []
            for i in item:
                r = self._convert_to_ids(i, key, meta, padding, return_text, *args, **kwargs)
                ret["seq_idx"].append(r["seq_idx"])
                ret["meta_idx"].append(r["meta_idx"])
                if return_text:
                    ret["seq_token"].append(r["seq_token"])
                    ret["meta"].append(r["meta"])
        else:
            ret = self._convert_to_ids(item, key, meta, padding, return_text, *args, **kwargs)

        return ret

    def _convert_to_ids(self, item: Union[str, dict, list], key=lambda x: x,
                        meta: Optional[list] = None,
                        padding=False, return_text=False, *args, **kwargs):
        token_item = self.tokenize(item, key)
        token_idx = []
        for _, w in enumerate(token_item):
            if isinstance(w, FigureSegment):
                # image
                try:
                    im = Image.open(os.path.join(self.img_dir, f'{w.src[10:-1]}.png'))
                    im = im.resize((56, 56))
                    token_idx.append(to_grayscale(im))
                except Exception:
                    warnings.warn('Open image error!')
                    token_idx.append(self.stoi['word'][self.img_token])
            else:
                # word
                token_idx.append(self.stoi['word'].get(w) or self.stoi['word'][self.unk_token])

        meta_idxs = {}
        meta_items = {}
        if meta is None:
            meta = self.meta
        for m in meta:
            meta_idx = []
            if isinstance(item, dict) and m in item:
                meta_item = item[m]
                if isinstance(meta_item, str):
                    if meta_item.startswith('['):
                        # a list of labels ['+', '-', '/']
                        meta_item = eval(meta_item)
                elif isinstance(meta_item, list):
                    pass
                else:
                    raise Exception("Side information must be a list!")
                meta_items[m] = meta_item
                for k in meta_item:
                    meta_idx.append(self.stoi[m].get(k) or self.stoi[m][self.unk_token])
            meta_idxs[m] = meta_idx
        ret = {
            "seq_idx": self.padding(token_idx, self.max_length) if padding else token_idx,
            "meta_idx": meta_idxs
        }

        if return_text:
            ret["seq_token"] = token_item
            ret["meta"] = meta_items
        return ret

    def load_meta_vocab(self, meta_vocab_dir):
        for m in self.meta:
            try:
                with open(os.path.join(meta_vocab_dir, f'meta_{m}.txt'), "rt", encoding="utf-8") as f:
                    meta = f.read().strip().split('\n')
                    self.stoi[m] = {word: index for index, word in enumerate(meta)}
                    self.itos[m] = {i: s for s, i in self.stoi[m].items()}
            except Exception:
                self.stoi[m] = None
                self.itos[m] = None

    def set_meta_vocab(self, items: list, meta: List[str] = None, silent=True):
        self.meta = meta if meta is not None else self.meta
        for m in self.meta:
            meta = set()
            if m in items[0]:
                for item in items:
                    meta_item = item[m]
                    if isinstance(meta_item, str):
                        if meta_item.startswith('['):
                            # a list of labels ['+', '-', '/']
                            meta_item = eval(meta_item)
                    elif isinstance(meta_item, list):
                        pass
                    else:
                        raise Exception("Side information must be a list!")
                    meta = meta | set(meta_item)
            meta = [self.unk_token] + sorted(meta)
            if not silent:
                print(f"save meta information {m}: {len(meta)}")
            self.stoi[m] = {word: index for index, word in enumerate(meta)}
            self.itos[m] = {i: s for s, i in self.stoi[m].items()}

    def set_vocab(self, items: list, key=lambda x: x, lower: bool = False,
                  trim_min_count: int = 1, do_tokenize: bool = True, silent=True):
        """
        Parameters
        -----------
        items: list
            can be the list of str, or list of dict
        key: function
            determine how to get the text of each item
        trim_min_count : int, optional
            the lower bound number for adding a word into vocabulary, by default 1
        silent
        """
        token_items = self.tokenize(items, key) if do_tokenize else [key(item) for item in items]
        self.vocab.set_vocab(corpus_items=token_items, trim_min_count=trim_min_count, lower=lower, silent=silent)
        self.stoi['word'] = self.vocab.token_to_idx
        self.itos['word'] = self.vocab.idx_to_token

    @classmethod
    def from_pretrained(cls, tokenizer_config_dir, img_dir=None, **kwargs):
        """
        Parameters:
        -----------
        tokenizer_config_dir: str
            must contain tokenizer_config.json and vocab.txt and meta_{meta_name}.txt
        img_dir: str
            default None
            the path of image directory
        """
        tokenizer_config_path = os.path.join(tokenizer_config_dir, "tokenizer_config.json")
        pretrained_vocab_path = os.path.join(tokenizer_config_dir, "vocab.txt")
        with open(tokenizer_config_path, "r", encoding="utf-8") as rf:
            tokenizer_config = json.load(rf)
            tokenizer_config.update(kwargs)
            return cls(
                vocab_path=pretrained_vocab_path,
                meta_vocab_dir=tokenizer_config_dir,
                img_dir=img_dir,
                max_length=tokenizer_config["max_length"],
                tokenize_method=tokenizer_config["tokenize_method"],
                meta=tokenizer_config["meta"],
                img_token=tokenizer_config["img_token"],
                unk_token=tokenizer_config["unk_token"],
                pad_token=tokenizer_config["pad_token"])

    def save_pretrained(self, tokenizer_config_dir):
        """Save tokenizer into local files

        Parameters:
        -----------
        tokenizer_config_dir: str
            save tokenizer params in `/tokenizer_config.json` and save words in `vocab.txt`
            and save metas in `meta_{meta_name}.txt`
        """
        if not os.path.exists(tokenizer_config_dir):
            os.makedirs(tokenizer_config_dir, exist_ok=True)
        tokenizer_config_path = os.path.join(tokenizer_config_dir, "tokenizer_config.json")
        self.vocab.save_vocab(os.path.join(tokenizer_config_dir, "vocab.txt"))
        for m in self.meta:
            save_list(self.stoi[m], os.path.join(tokenizer_config_dir, f'meta_{m}.txt'))
        with open(tokenizer_config_path, "w", encoding="utf-8") as wf:
            json.dump(self.config, wf, ensure_ascii=False, indent=2)

    def padding(self, idx, max_length, type='word'):
        padding_idx = idx + [self.stoi[type][self.pad_token]] * (max_length - len(idx))
        return padding_idx

    def set_img_dir(self, path):
        self.img_dir = path


def clip(v, low, high):
    if v < low:
        v = low
    if v > high:
        v = high
    return v


class Lines:
    def __init__(self, filename, skip=0, preserve_newline=False):
        self.filename = filename
        with open(filename, "r", encoding="utf-8") as f:
            self.length = len(f.readlines()) - skip
        assert self.length > 0, f'{filename} is empty. Or file length is less than skip length.'
        self.skip = skip
        self.preserve_newline = preserve_newline

    def __len__(self):
        return self.length

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, item):
        d = self.skip + 1
        if isinstance(item, int):
            if item < self.length:
                line = linecache.getline(self.filename,
                                         item % len(self) + d)
                if self.preserve_newline:
                    return json.loads(line)
                else:
                    return json.loads(line.strip('\r\n'))

        elif isinstance(item, slice):
            low = 0 if item.start is None else item.start
            low = clip(low, -len(self), len(self) - 1)
            if low < 0:
                low += len(self)
            high = len(self) if item.stop is None else item.stop
            high = clip(high, -len(self), len(self))
            if high < 0:
                high += len(self)
            ls = []
            for i in range(low, high):
                line = linecache.getline(self.filename, i + d)
                if not self.preserve_newline:
                    line = line.strip('\r\n')
                ls.append(json.loads(line))

            return ls

        raise IndexError('index must be int or slice')


class QuestionLoader:
    def __init__(self, ques: Lines, tokenizer: QuesNetTokenizer,
                 pipeline=None, range=None, meta: Optional[list] = None,
                 content_key=lambda x: x['ques_content'],
                 meta_key=lambda x: x['know_name'],
                 answer_key=lambda x: x['ques_answer'],
                 option_key=lambda x: x['ques_options'],
                 skip=0
                 ):
        """ Read question file as data list. Same behavior on same file.

        Parameters
        ----------
        ques_file : str
            path of question file
        tokenizer : QuesNetTokenizer
        pipeline : _type_, optional
            _description_, by default None
        range : _type_, optional
            _description_, by default None
        content_key : function, optional
            by default lambda x:x['ques_content']
        meta_key : function, optional
            by default lambda x:x['know_name']
        answer_key: function, optional
            by default lambda x:x['ques_answer']
        option_key: function, optional
            by default lambda x:x['ques_options']
        skip: int, optional
            skip the first several lines, by default 0
        """
        self.range = None
        self.ques = ques
        self.range = range or slice(0, len(self), skip)
        self.img_dir = tokenizer.img_dir
        self.labels = []
        self.stoi = tokenizer.stoi
        self.tokenizer = tokenizer

        self.content_key = content_key
        self.meta = meta if meta else tokenizer.meta
        self.meta_key = meta_key
        self.answer_key = answer_key
        self.option_key = option_key

        self.pipeline = pipeline

    def split_(self, split_ratio):
        first_size = int(len(self) * (1 - split_ratio))
        other = copy(self)
        self.range = slice(0, first_size, 1)
        other.range = slice(first_size, len(other), 1)
        return other

    def __len__(self):
        return len(self.ques) if self.range is None \
            else self.range.stop - self.range.start

    def __getitem__(self, x):
        if isinstance(x, int):
            x += self.range.start
            item = slice(x, x + 1, 1)
        else:
            item = slice(x.start + self.range.start,
                         x.stop + self.range.start, 1)
        qs = []
        if item.start > len(self):
            raise IndexError
        for line in self.ques[item]:
            q = line
            qid = q['ques_id']
            token = self.tokenizer(q, key=self.content_key, meta=self.meta)
            content = token['seq_idx']
            meta = token['meta_idx']
            if self.answer_key(q).isalpha() and len(self.answer_key(q)) == 1 and ord(self.answer_key(q)) < 128 and len(
                    self.option_key(q)) > 0:
                answer_idx = ord(self.answer_key(q).upper()) - ord('A')
                options = self.option_key(q)
                answer = self.tokenizer(options.pop(answer_idx), meta=self.meta)
                answer = answer['seq_idx']
                false_options = [(self.tokenizer(option, meta=self.meta))['seq_idx'] for option in options]
                qs.append(Question(qid, content, answer, false_options, meta))
            else:
                answer = (self.tokenizer(self.answer_key(q), meta=self.meta))['seq_idx']
                qs.append(Question(qid, content, answer, [[0], [0], [0]], meta))

        if callable(self.pipeline):
            qs = self.pipeline(qs)
        if isinstance(x, int):
            return qs[0]
        else:
            return qs


def optimizer(*models, **kwargs):
    _cur_optim = [m.optim_cls(m.parameters(), **kwargs)
                  if hasattr(m, 'optim_cls')
                  else torch.optim.Adam(m.parameters(), **kwargs)
                  for m in models]
    if len(_cur_optim) == 1:
        return _cur_optim[0]
    else:
        return _cur_optim


class PrefetchIter:
    """Iterator on data and labels, with states for save and restore."""

    def __init__(self, data, *label, length=None, batch_size=1, shuffle=True):
        self.data = data
        self.label = label
        self.batch_size = batch_size
        self.queue = queue.Queue(maxsize=8)
        self.length = length if length is not None else len(data)

        assert all(self.length == len(lab) for lab in label), \
            'data and label must have same lengths'

        self.index = list(range(len(self)))
        if shuffle:
            random.shuffle(self.index)
        self.thread = None
        self.pos = 0

    def __len__(self):
        return math.ceil(self.length / self.batch_size)

    def __iter__(self):
        return self

    def __next__(self):
        if self.thread is None:
            self.thread = threading.Thread(target=self.produce, daemon=True)
            self.thread.start()

        if self.pos >= len(self.index):
            raise StopIteration

        item = self.queue.get()
        if isinstance(item, Exception):
            raise item
        else:
            self.pos += 1
            return item

    def produce(self):
        for i in range(self.pos, len(self.index)):
            try:
                index = self.index[i]

                bs = self.batch_size

                if callable(self.data):
                    data_batch = self.data(index * bs, (index + 1) * bs)
                else:
                    data_batch = self.data[index * bs:(index + 1) * bs]

                label_batch = [label[index * bs:(index + 1) * bs]
                               for label in self.label]
                if label_batch:
                    self.queue.put([data_batch] + label_batch)
                else:
                    self.queue.put(data_batch)
            except Exception as e:
                self.queue.put(e)
                return


class EmbeddingDataset(Dataset):
    def __init__(self, data, data_type='image'):
        self.data = data
        self.data_type = data_type
        assert self.data_type in ['image', 'meta']

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if self.data_type == 'image':
            return to_tensor(self.data[idx])
        elif self.data_type == 'meta':
            return self.data[idx]


def pretrain_iter(ques, batch_size):
    _cur_iter = PrefetchIter(ques, batch_size=batch_size)
    return _cur_iter


sigint_handler = signal.getsignal(signal.SIGINT)


def critical(f):
    it = iter(f)
    signal_received = ()

    def handler(sig, frame):
        nonlocal signal_received
        signal_received = (sig, frame)
        logging.debug('SIGINT received. Delaying KeyboardInterrupt.')

    while True:
        try:
            signal.signal(signal.SIGINT, handler)
            yield next(it)
            signal.signal(signal.SIGINT, sigint_handler)
            if signal_received:
                sigint_handler(*signal_received)
        except StopIteration:
            break


def pretrain_embedding_layer(dataset: EmbeddingDataset, ae: AE, lr: float = 1e-3, log_step: int = 1, epochs: int = 3,
                             batch_size: int = 4, device=torch.device('cpu')):
    train_dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
    optim = torch.optim.Adam(ae.parameters(), lr=lr)
    ae.train()
    ae.to(device)
    train_type = dataset.data_type
    for i in range(epochs):
        for batch, item in tqdm(enumerate(train_dataloader)):
            loss = ae.loss(item.to(device))
            optim.zero_grad()
            loss.backward()
            optim.step()
            if batch % log_step == 0:
                logger.info(f"[Epoch{i}][Batch{batch}]Training {train_type} Embedding layer, loss:{loss}")
    return ae


def pretrain_quesnet(path, output_dir, img_dir=None, save_embs=False, train_params=None):
    """ pretrain quesnet

    Parameters
    ----------
    path : str
        path of question file
    output_dir : str
        output path·
    tokenizer : QuesNetTokenizer
        quesnet tokenizer
    save_embs : bool, optional
        whether to save pretrained word/image/meta embeddings seperately
    train_params : dict, optional
        the training parameters and model parameters, by default None
        - "n_epochs": int, default = 1
            train param, number of epochs
        - "batch_size": int, default = 6
            train param, batch size
        - "lr": float, default = 1e-3
            train param, learning rate
        - "save_every": int, default = 0
            train param, save steps interval
        - "log_steps": int, default = 10
            train param, log steps interval
        - "device": str, default = 'cpu'
            train param, 'cpu' or 'cuda'
        - "max_steps": int, default = 0
            train param, stop training when reach max steps
        - "emb_size": int, default = 256
            model param, the embedding size of word, figure, meta info
        - "feat_size": int, default = 256
            model param, the size of question infer vector

    Examples
    ----------
    >>> tokenizer = QuesNetTokenizer(meta=['know_name'])
    >>> items = [{"ques_content": "若复数$z=1+2 i+i^{3}$，则$|z|=$，$\\FigureID{000004d6-0479-11ec-829b-797d5eb43535}$",
    ... "ques_id": "726cdbec-33a9-11ec-909c-98fa9b625adb",
    ... "know_name": "['代数', '集合', '集合的相等']"
    ... }]
    >>> tokenizer.set_vocab(items, key=lambda x: x['ques_content'], trim_min_count=1, silent=True)
    >>> pretrain_quesnet('./data/standard_luna_data.json', './testQuesNet', tokenizer) # doctest: +SKIP
    """
    default_train_params = {
        # train params
        "n_epochs": 1,
        "batch_size": 8,
        "lr": 1e-3,
        'save_every': 0,
        'log_steps': 10,
        'device': 'cpu',
        'max_steps': 0,
        # model params
        'emb_size': 256,
        'feat_size': 256,
    }
    if train_params is not None:
        default_train_params.update(train_params)
    train_params = default_train_params
    os.makedirs(output_dir, exist_ok=True)
    device = torch.device(train_params['device'])
    # set tokenizer
    tokenizer = QuesNetTokenizer(meta=['know_name'],
                                 img_dir=img_dir)
    items = Lines(path, skip=1)

    tokenizer.set_vocab(items, key=lambda x: x['ques_content'],
                        trim_min_count=2, silent=False)
    tokenizer.set_meta_vocab(items, silent=False)
    tokenizer.save_pretrained(output_dir)

    ques_dl = QuestionLoader(items, tokenizer)
    model = QuesNetForPreTraining(_stoi=tokenizer.stoi, feat_size=train_params['feat_size'],
                                  emb_size=train_params['emb_size']).to(device)
    emb_dict = tokenizer.stoi['word']
    emb_dict_rev = tokenizer.itos['word']
    emb_size = train_params['emb_size']
    meta_size = model.quesnet.meta_size
    w2v_corpus = []
    img_corpus = []
    meta_corpus = []
    for i, qs in enumerate(tqdm(ques_dl)):
        text_content = []
        for c in qs.content:
            if isinstance(c, int):
                text_content.append(emb_dict_rev[c])
            else:
                img_corpus.append(c)
        for a in qs.answer:
            if isinstance(a, int):
                text_content.append(emb_dict_rev[a])
            else:
                img_corpus.append(a)
        w2v_corpus.append(text_content)
        meta_vector = torch.zeros(meta_size, dtype=torch.float)
        for m in qs.labels[model.quesnet.meta]:
            meta_vector.add_(torch.nn.functional.one_hot(torch.tensor(m, dtype=torch.int64),
                             model.quesnet.meta_size).to(torch.float))
        meta_corpus.append(meta_vector)

    # train word2vec for text embedding
    gensim_w2v = Word2Vec(sentences=[[item] for item in emb_dict.keys()], min_count=1,
                          vector_size=emb_size)
    gensim_w2v.init_weights()
    gensim_w2v.train(corpus_iterable=w2v_corpus, total_examples=len(w2v_corpus), epochs=train_params['n_epochs'])
    w2v_emb = gensim_w2v.syn1neg
    emb_weights = []
    for key, item in emb_dict.items():
        w2v_index = gensim_w2v.wv.key_to_index[key]
        emb_weights.append(w2v_emb[w2v_index])
    emb_weights = np.array(emb_weights)
    model.quesnet.load_emb(emb_weights)
    logger.info('quesnet Word Embedding loaded')
    if save_embs:
        np.save(os.path.join(output_dir, 'w2v_embs.npy'), emb_weights)

    # train auto-encoder loss for image embedding
    img_dataset = EmbeddingDataset(data=img_corpus, data_type='image')
    trained_ie = pretrain_embedding_layer(dataset=img_dataset, ae=model.quesnet.ie, lr=train_params['lr'],
                                          log_step=train_params['log_steps'], batch_size=train_params['batch_size'],
                                          epochs=train_params['n_epochs'], device=device)
    model.quesnet.load_img(trained_ie)
    logger.info('quesnet Image Embedding loaded')
    if save_embs:
        torch.save(trained_ie.state_dict(), os.path.join(output_dir, 'trained_ie.pt'))

    # train auto-encoder loss for meta embedding
    meta_dateset = EmbeddingDataset(data=meta_corpus, data_type='meta')
    trained_me = pretrain_embedding_layer(dataset=meta_dateset, ae=model.quesnet.me, lr=train_params['lr'],
                                          log_step=train_params['log_steps'], batch_size=train_params['batch_size'],
                                          epochs=train_params['n_epochs'], device=device)
    model.quesnet.load_meta(trained_me)
    logger.info('quesnet Meta Embedding loaded')
    if save_embs:
        torch.save(trained_me.state_dict(), os.path.join(output_dir, 'trained_me.pt'))

    logger.info("quesnet Word, Image and Meta Embeddings training is done")

    # HLM and DOO training
    ques_dl.pipeline = partial(model.quesnet.make_batch, device=device, pretrain=True)
    model.train()
    optim = optimizer(model, lr=train_params['lr'])
    n_batches = 0
    for epoch in range(0, train_params['n_epochs']):
        train_iter = pretrain_iter(ques_dl, train_params['batch_size'])
        bar = enumerate(tqdm(train_iter, initial=train_iter.pos),
                        train_iter.pos)
        for i, batch in critical(bar):
            n_batches += 1
            loss = model(batch).loss
            optim.zero_grad()
            loss.backward()
            optim.step()

            if train_params['save_every'] > 0 and i % train_params['save_every'] == 0:
                # model.save(os.path.join(output_dir, f'QuesNet_{epoch}.{i}'))
                model.save_pretrained(output_dir)

            if train_params['log_steps'] > 0 and i % train_params['log_steps'] == 0:
                logger.info(f"{epoch}.{i}---loss: {loss.item()}")

            if train_params['max_steps'] > 0 and n_batches % train_params['max_steps'] == 0:
                break

        # model.save(os.path.join(output_dir, f'QuesNet_{epoch}'))

    model.save_pretrained(output_dir)
    tokenizer.save_pretrained(output_dir)



# -*- coding: utf-8 -*-

import torch
from torch import nn
from torch.nn import functional as F


def get_mask(seq_len, lengths):
    device = lengths.device
    # batch_size
    batch_size = lengths.size(0)
    # seq_len
    pos_index = torch.arange(seq_len).to(device)
    # batch_size * seq_len
    mask = pos_index.unsqueeze(0).expand(batch_size, -1) >= lengths.unsqueeze(-1)
    return mask


def shuffle(real):
    # |0 1 2 3| => |1 2 3 0|
    device = real.device
    batch_size = real.size(0)
    shuffled_index = (torch.arange(batch_size) + 1) % batch_size
    shuffled_index = shuffled_index.to(device)
    shuffled = real.index_select(dim=0, index=shuffled_index)
    return shuffled


def spectral_norm(w, n_iteration=5):
    device = w.device
    # (o, i)
    # bias: (O) -> (o, 1)
    if w.dim() == 1:
        w = w.unsqueeze(-1)
    out_dim, in_dim = w.size()
    # (i, o)
    wt = w.transpose(0, 1)
    # (1, i)
    u = torch.ones(1, in_dim).to(device)
    for _ in range(n_iteration):
        # (1, i) * (i, o) -> (1, o)
        v = torch.mm(u, wt)
        v = v / v.norm(p=2)
        # (1, o) * (o, i) -> (1, i)
        u = torch.mm(v, w)
        u = u / u.norm(p=2)
    # (1, i) * (i, o) * (o, 1) -> (1, 1)
    sn = torch.mm(torch.mm(u, wt), v.transpose(0, 1)).sum() ** 0.5
    return sn



# from transformers import BertModel as HFBertModel
from transformers import AutoModel
from .const import UNK, PAD
from .meta import Vector
import torch


class BertModel(Vector):
    """
    Examples
    --------
    >>> from EduNLP.Pretrain import BertTokenizer
    >>> tokenizer = BertTokenizer("bert-base-chinese", add_special_tokens=False)
    >>> model = BertModel("bert-base-chinese")
    >>> item = ["有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$，若$x,y$满足约束",
    ... "有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$，若$x,y$满足约束"]
    >>> inputs = tokenizer(item, return_tensors='pt')
    >>> output = model(inputs)
    >>> output.shape
    torch.Size([2, 14, 768])
    >>> tokens = model.infer_tokens(inputs)
    >>> tokens.shape
    torch.Size([2, 12, 768])
    >>> tokens = model.infer_tokens(inputs, return_special_tokens=True)
    >>> tokens.shape
    torch.Size([2, 14, 768])
    >>> item = model.infer_vector(inputs)
    >>> item.shape
    torch.Size([2, 768])
    """

    def __init__(self, pretrained_model):
        self.model = AutoModel.from_pretrained(pretrained_model)

    def __call__(self, items: dict):
        # batch_size, sent_len, embedding_size
        tokens = self.model(**items).last_hidden_state
        return tokens

    def infer_vector(self, items: dict, pooling_strategy='CLS') -> torch.Tensor:
        vector = self(items)
        if pooling_strategy == 'CLS':
            return vector[:, 0, :]
        elif pooling_strategy == 'average':
            # the average of word embedding of the last layer
            # batch_size, sent_len, embedding_dim
            mask = items['attention_mask'].unsqueeze(-1).expand(vector.size())
            mul_mask = vector * mask
            # batch_size, embedding_dim
            return mul_mask.sum(1) / (mask.sum(1) + 1e-10)

    def infer_tokens(self, items: dict, return_special_tokens=False) -> torch.Tensor:
        tokens = self(items)
        if return_special_tokens:
            # include embedding of [CLS] and [SEP]
            return tokens
        else:
            # ignore embedding of [CLS] and [SEP]
            return tokens[:, 1:-1, :]

    @property
    def vector_size(self):
        return self.model.config.hidden_size



import torch
from typing import Union
from EduNLP.ModelZoo.quesnet import QuesNet
from EduNLP.Pretrain import Question, QuesNetTokenizer


class QuesNetModel(object):
    def __init__(self, pretrained_dir, img_dir=None, device="cpu", **kwargs):
        """
        Parameters
        ----------
        pretrained_dir: str
            the dirname to pretrained model
        device: str
            cpu or cuda, default is cpu
        img_dir: str
            image dir
        """
        self.device = torch.device(device)
        self.model = QuesNet.from_pretrained(pretrained_dir, img_dir=img_dir).to(device)
        self.model.eval()

    def __call__(self, items: dict, **kwargs):
        """ get question embedding with quesnet

        Parameters
        ----------
        items:
            encodes from tokenizer
        """
        qs = [Question("", items['seq_idx'][i],
                       [0], [[0], [0], [0]], items['meta_idx'][i]) for i in range(len(items['seq_idx']))]
        outputs = self.model(self.model.make_batch(qs, device=self.device))
        return outputs.hidden, outputs.embeded

    def infer_vector(self, items: Union[dict, list]) -> torch.Tensor:
        """ get question embedding with quesnet

        Parameters
        ----------
        items:
            encodes from tokenizer
        """
        return self(items)[0]

    def infer_tokens(self, items: Union[dict, list]) -> torch.Tensor:
        """ get token embeddings with quesnet

        Parameters
        ----------
        items:
            encodes from tokenizer
        Returns
        -------
        torch.Tensor
            word_embs + meta_emb
        """
        vector = self(items)[1]
        """ Please note that output vector is like 0 0 seq_idx(text with image) 0 meta_idx 0 0"""
        return vector[:, 2:-2, :]

    @property
    def vector_size(self):
        return self.model.feat_size



# coding: utf-8
# 2021/5/29 @ tongshiwei

import numpy as np
from pathlib import PurePath
from gensim.models import KeyedVectors, Word2Vec, FastText, Doc2Vec, TfidfModel
from gensim import corpora
import re
from .const import UNK, PAD
from .meta import Vector


class W2V(Vector):
    """
    The part uses gensim library providing FastText, Word2Vec and KeyedVectors method to transfer word to vector.

    Parameters
    ----------
    filepath:
        path to the pretrained model file
    method: str
        fasttext
        other(Word2Vec)
    binary: bool
    """
    def __init__(self, filepath, method=None, binary=None):
        fp = PurePath(filepath)
        self.binary = binary if binary is not None else (True if fp.suffix == ".bin" else False)
        if self.binary is True:
            if method == "fasttext":
                self.wv = FastText.load(filepath).wv
            else:
                self.wv = Word2Vec.load(filepath).wv
        else:
            self.wv = KeyedVectors.load(filepath, mmap="r")

        self.method = method
        self.constants = {UNK: 0, PAD: 1}

    def __len__(self):
        return len(self.constants) + len(self.wv.key_to_index)

    def key_to_index(self, word):
        if word in self.constants:
            return self.constants[word]
        else:
            if word in self.wv.key_to_index:
                return self.wv.key_to_index[word] + len(self.constants)
            else:
                return self.constants[UNK]

    @property
    def vectors(self):
        return np.concatenate([np.zeros((len(self.constants), self.vector_size)), self.wv.vectors], axis=0)

    @property
    def vector_size(self):
        return self.wv.vector_size

    def __call__(self, *words):
        for word in words:
            yield self[word]

    def __getitem__(self, item):
        index = self.key_to_index(item)
        return self.wv[item] if index not in self.constants.values() else np.zeros((self.vector_size,))

    def infer_vector(self, items, agg="mean", *args, **kwargs) -> list:
        token_vectors = self.infer_tokens(items, *args, **kwargs)
        return [eval("np.%s" % agg)(item, axis=0) for item in token_vectors]

    def infer_tokens(self, items, *args, **kwargs) -> list:
        return [list(self(*item)) for item in items]


class BowLoader(object):
    """
    Using doc2bow model, which has a lot of effects.

    Convert document (a list of words) into the bag-of-words format = list of \
    (token_id, token_count) 2-tuples. Each word is assumed to be a \
    tokenized and normalized string (either unicode or utf8-encoded). \
    No further preprocessing is done on the words in document;\
     apply tokenization, stemming etc. before calling this method.

    If allow_update is set, then also update dictionary in the process: \
    create ids for new words. At the same time, update document frequencies – \
    for each word appearing in this document, increase its document frequency (self.dfs) by one.

    If allow_update is not set, this function is const, \
    aka read-only.
    """
    def __init__(self, filepath):
        self.dictionary = corpora.Dictionary.load(filepath)

    def infer_vector(self, item, return_vec=False):
        item = self.dictionary.doc2bow(item)
        if not return_vec:
            return item  # return dic as default
        vec = [0 for i in range(len(self.dictionary.keys()))]
        for i, v in item:
            vec[i] = v
        return vec

    @property
    def vector_size(self):
        return len(self.dictionary.keys())


class TfidfLoader(object):
    """
    This module implements functionality related to the Term Frequency - \
    Inverse Document Frequency <https://en.wikipedia.org/wiki/Tf%E2%80%93idf> \
    vector space bag-of-words models.
    """
    def __init__(self, filepath):
        self.tfidf_model = TfidfModel.load(filepath)
        # 'tfidf' model shold be used based on 'bow' model
        dictionary_path = re.sub(r"(.*)tfidf", r"\1bow", filepath)
        self.dictionary = corpora.Dictionary.load(dictionary_path)

    def infer_vector(self, item, return_vec=False):
        dic_item = self.dictionary.doc2bow(item)
        tfidf_item = self.tfidf_model[dic_item]
        # return dic as default
        if not return_vec:
            return tfidf_item  # pragma: no cover
        vec = [0 for i in range(len(self.dictionary.keys()))]
        for i, v in tfidf_item:
            vec[i] = v
        return vec

    @property
    def vector_size(self):
        return len(self.dictionary.token2id)


class D2V(Vector):
    """
    It is a collection which include d2v, bow, tfidf method.

    Parameters
    -----------
    filepath
    method: str
        d2v
        bow
        tfidf
    item

    Returns
    ---------
    d2v model:D2V
    """
    def __init__(self, filepath, method="d2v"):
        self._method = method
        self._filepath = filepath
        if self._method == "d2v":
            self.d2v = Doc2Vec.load(filepath)
        elif self._method == "bow":
            self.d2v = BowLoader(filepath)
        elif self._method == "tfidf":
            self.d2v = TfidfLoader(filepath)
        else:
            raise ValueError("Unknown method: %s" % method)

    def __call__(self, item):
        if self._method == "d2v":
            return self.d2v.infer_vector(item)
        else:
            return self.d2v.infer_vector(item, return_vec=True)

    @property
    def vector_size(self):
        if self._method == "d2v":
            return self.d2v.vector_size
        elif self._method == "bow":
            return self.d2v.vector_size
        elif self._method == "tfidf":
            return self.d2v.vector_size

    def infer_vector(self, items, *args, **kwargs) -> list:
        return [self(item) for item in items]

    def infer_tokens(self, item, *args, **kwargs) -> ...:
        raise NotImplementedError



import os
import numpy as np
import torch
from copy import deepcopy
from transformers import TrainingArguments, Trainer
import warnings
from .pretrian_utils import EduDataset
from typing import Dict, List, Tuple, Union, Any
from gensim.models import Word2Vec
from torch.optim.adam import Adam
from torch.optim.lr_scheduler import StepLR
import torch.nn as nn
from dataclasses import dataclass, field
from ..SIF import EDU_SPYMBOLS
from ..ModelZoo.disenqnet.disenqnet import DisenQNetForPreTraining
from ..ModelZoo.utils import load_items, pad_sequence
from .pretrian_utils import PretrainedEduTokenizer


def check_num(s):
    # (1/2) -> 1/2
    if s.startswith('(') and s.endswith(')'):
        if s == "()":
            return False
        else:
            s = s[1:-1]
    # -1/2 -> 1/2
    if s.startswith('-') or s.startswith('+'):
        if s == '-' or s == '+':
            return False
        else:
            s = s[1:]
    # 1/2
    if '/' in s:
        if s == '/':
            return False
        else:
            sep_index = s.index('/')
            is_num = s[:sep_index].isdigit() and s[sep_index + 1:].isdigit()
            return is_num
    else:
        # 1.2% -> 1.2
        if s.endswith('%'):
            if s == '%':
                return False
            else:
                s = s[:-1]
        # 1.2
        if '.' in s:
            try:
                float(s)
                return True
            except ValueError:
                return False
        # 12
        else:
            is_num = s.isdigit()
            return is_num


def load_list_to_dict(path):
    with open(path, "rt", encoding="utf-8") as file:
        items = file.read().split('\n')
    item2index = {item: index for index, item in enumerate(items)}
    return item2index


def save_dict_to_list(item2index, path):
    item2index = sorted(item2index.items(), key=lambda kv: kv[1])
    items = [item for item, _ in item2index]
    with open(path, "wt", encoding="utf-8") as file:
        file.write('\n'.join(items))
    return


class DisenQTokenizer(PretrainedEduTokenizer):
    """
    Examples
    --------
    >>> tokenizer = DisenQTokenizer()
    >>> test_items = [{
    ...     "content": "甲 数 除以 乙 数 的 商 是 1.5 ， 如果 甲 数 增加 20 ， 则 甲 数 是 乙 的 4 倍 ． 原来 甲 数 = ．",
    ...     "knowledge": ["*", "-", "/"], "difficulty": 0.2, "length": 7}]
    >>> tokenizer.set_vocab(test_items,
    ...     key=lambda x: x["content"], trim_min_count=1)
    [['甲', '数', '除以', '乙', '数', '商', '[NUM]', '甲', '数', '增加', '[NUM]', '甲', '数', '乙', '倍', '甲', '数']]
    >>> token_items = [tokenizer(i, key=lambda x: x["content"]) for i in test_items]
    >>> print(token_items[0].keys())
    dict_keys(['seq_idx', 'seq_len'])
    """

    def __init__(self, vocab_path=None, max_length=250, tokenize_method="pure_text",
                 add_specials: list = None, num_token="[NUM]", **kwargs):
        """
        Parameters
        ----------
        vocab_path: str
            default is None
        max_length: int
            default is 250, used to clip the sentence out of length
        tokenize_method: str
            default: "pure_text"
            when text is already seperated by space, use "space"
            when text is raw string format, use Tokenizer defined in get_tokenizer(), such as "pure_text" and "text"
        num_token: str
        """
        if add_specials is None:
            add_specials = [num_token]
        else:
            add_specials = [num_token] + add_specials
        super().__init__(vocab_path=vocab_path, max_length=max_length,
                         tokenize_method=tokenize_method, add_specials=add_specials, **kwargs)
        self.num_token = num_token
        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__", "vocab_path"]}

    def _tokenize(self, item: Tuple[str, dict], key=lambda x: x):
        token_item = self.text_tokenizer._tokenize(item, key=key)
        if len(token_item) == 0:
            token_item = [self.vocab.unk_token]
        if len(token_item) > self.max_length:
            token_item = token_item[:self.max_length]
        token_item = [self.num_token if check_num(w) else w for w in token_item]
        return token_item


def preprocess_dataset(pretrained_dir, disen_tokenizer, items, data_formation, trim_min_count=None, embed_dim=None,
                       w2v_params=None, silent=False):
    default_w2v_params = {
        "workers": 1,
    }
    if w2v_params is not None:
        default_w2v_params.update(w2v_params)
    w2v_params = default_w2v_params

    concept_list_path = os.path.join(pretrained_dir, "concept.list")
    vocab_path = os.path.join(pretrained_dir, "vocab.list")
    wv_path = os.path.join(pretrained_dir, "wv.th")

    file_num = sum(map(lambda x: os.path.exists(x), [wv_path, concept_list_path]))
    if file_num > 0 and file_num < 3:
        warnings.warn("Some files are missed in pretrained_dir(which including wv.th, vocab.list, and concept.list)",
                      UserWarning)

    if not os.path.exists(vocab_path):
        # load word
        token_items = disen_tokenizer.set_vocab(items, key=lambda x: x[data_formation["ques_content"]],
                                                trim_min_count=trim_min_count)
        disen_tokenizer.save_pretrained(pretrained_dir)
        if not silent:
            print(f"save vocab to {vocab_path}")
    else:
        if not silent:
            print(f"load vocab from {vocab_path}")

    # construct concept list
    if not os.path.exists(concept_list_path):
        concepts = set()
        for data in items:
            print(data)
            concept = data[data_formation["knowledge"]]
            for c in concept:
                if c not in concepts:
                    concepts.add(c)
        concepts = sorted(concepts)
        concept_to_idx = {concept: index for index, concept in enumerate(concepts)}
        save_dict_to_list(concept_to_idx, concept_list_path)
        if not silent:
            print(f"save concept to {concept_list_path}")
    else:
        print(f"load concept from {concept_list_path}")
        concept_to_idx = load_list_to_dict(concept_list_path)

    # word2vec
    if not os.path.exists(wv_path):
        words = disen_tokenizer.vocab.tokens
        unk_token = disen_tokenizer.vocab.unk_token
        corpus = list()
        word_set = set(words)
        for text in token_items:
            text = [w if w in word_set else unk_token for w in text]
            corpus.append(text)
        wv = Word2Vec(corpus, vector_size=embed_dim, min_count=trim_min_count, **w2v_params).wv
        # 按照 vocab 中的词序 来保存
        wv_list = [wv[w] if w in wv.key_to_index else np.random.rand(embed_dim) for w in words]
        word2vec = torch.tensor(wv_list)
        torch.save(word2vec, wv_path)
        if not silent:
            print(f"save word2vec to {wv_path}")
    else:
        print(f"load word2vec from {wv_path}")
        word2vec = torch.load(wv_path)
    return disen_tokenizer, concept_to_idx, word2vec


class DisenQDataset(EduDataset):
    def __init__(self, items: List[Dict], tokenizer: DisenQTokenizer, data_formation: Dict,
                 mode="train", concept_to_idx=None, **kwargs):
        """
        Parameters
        ----------
        texts: list
        tokenizer: DisenQTokenizer
        data_formation: dict
        max_length: int, optional, default=128
        """
        # super(DisenQDataset, self).__init__(tokenizer=tokenizer, **kwargs)
        self.tokenizer = tokenizer
        self.concept_to_idx = concept_to_idx
        self.mode = mode
        self.items = items
        self.max_length = tokenizer.max_length
        self.data_formation = data_formation

    def __len__(self):
        return len(self.items)

    def _list_to_onehot(self, item_list, item2index):
        onehot = np.zeros(len(item2index)).astype(np.int64)
        for c in item_list:
            onehot[item2index[c]] = 1
        return onehot

    def __getitem__(self, index):
        item = self.items[index]
        ret = self.tokenizer(item, padding=False, key=lambda x: x[self.data_formation["ques_content"]],
                             return_tensors=False, return_text=False)
        if self.mode in ["train", "val"]:
            ret['concept'] = self._list_to_onehot(item[self.data_formation["knowledge"]], self.concept_to_idx)
        return ret

    def collate_fn(self, batch_data):
        pad_idx = self.tokenizer.vocab.pad_idx
        first = batch_data[0]
        batch = {
            k: [item[k] for item in batch_data] for k in first.keys()
        }
        batch["seq_idx"] = pad_sequence(batch["seq_idx"], pad_val=pad_idx)
        batch = {key: torch.as_tensor(val) for key, val in batch.items()}
        return batch


class DisenQTrainer(Trainer):
    def create_optimizer_and_scheduler(self, num_training_steps: int):
        self.optimizer = Adam(self.model.model_params, lr=self.args.learning_rate)
        self.lr_scheduler = StepLR(self.optimizer, step_size=self.args.step_size, gamma=self.args.gamma)
        self.adv_optimizer = Adam(self.model.adv_params, lr=self.args.learning_rate)
        self.adv_scheduler = StepLR(self.adv_optimizer, step_size=self.args.step_size, gamma=self.args.gamma)

    def training_step(self, model: nn.Module, inputs: Dict[str, Union[torch.Tensor, Any]]) -> torch.Tensor:
        model.train()
        inputs = self._prepare_inputs(inputs)
        warming_up = self.state.epoch <= model.params["warmup"]
        print(self.state)
        if not warming_up:
            # train disc
            outputs = model(**inputs)
            # stop gradient propagation to encoder
            k_hidden = outputs.k_hidden.detach()
            i_hidden = outputs.i_hidden.detach()
            # max dis_loss
            dis_loss = - model.disen_estimator(k_hidden, i_hidden)
            dis_loss = model.params["n_adversarial"] * model.params["w_dis"] * dis_loss
            if self.args.gradient_accumulation_steps > 1 and not self.deepspeed:
                dis_loss = dis_loss / self.args.gradient_accumulation_steps
            dis_loss.backward()
            step = self.state.global_step % (self.state.max_steps / self.state.num_train_epochs)
            if (step + 1) % self.args.gradient_accumulation_steps or \
                    (step + 1) == self.state.max_steps / self.state.num_train_epochs:
                self.adv_optimizer.step()
                self.adv_optimizer.zero_grad()
            # Lipschitz constrain for Disc of WGAN
            model.disen_estimator.spectral_norm()
        model.warming_up = warming_up

        loss = self.compute_loss(model, inputs)

        if self.args.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training

        if self.args.gradient_accumulation_steps > 1 and not self.deepspeed:
            # deepspeed handles loss scaling by gradient_accumulation_steps in its `backward`
            loss = loss / self.args.gradient_accumulation_steps

        loss.backward()

        # TODO: Not sure. if warming_up, the scheduler should not step?
        if not warming_up:
            self.adv_scheduler.step()

        return loss.detach()


@dataclass
class DisenQTrainingArguments(TrainingArguments):
    step_size: int = field(default=False, metadata={"help": "step_size"})
    trim_min: int = field(default=False, metadata={"help": "trim"})
    hidden_size: int = field(default=False, metadata={"help": "hidden_size"})
    gamma: float = field(default=False, metadata={"help": "gamma"})


DEFAULT_TRAIN_PARAMS = {
    # default
    "output_dir": None,
    "overwrite_output_dir": True,
    "num_train_epochs": 3,
    "per_device_train_batch_size": 8,
    "per_device_eval_batch_size": 32,
    # evaluation_strategy: "steps",
    # eval_steps:200,
    "save_steps": 1000,
    "save_total_limit": 2,
    # "load_best_model_at_end": False,
    # metric_for_best_model: "loss",
    # greater_is_better: False,
    "logging_dir": None,
    "logging_steps": 5,
    "gradient_accumulation_steps": 1,
    "learning_rate": 5e-4,
    # disable_tqdm: True,
    # no_cuda: True,
    'step_size': 20,
    'gamma': 0.5,
    "trim_min": 1,
    "hidden_size": 300
}


def train_disenqnet(train_items: List[dict], output_dir: str, pretrained_dir: str = None,
                    eval_items=None, tokenizer_params=None, data_params=None, model_params=None,
                    train_params=None, w2v_params=None):
    """
    Parameters
    ----------
    train_items : List[dict]
        _description_
    output_dir : str
        _description_
    pretrained_dir : str, optional
        _description_, by default None
    tokenizer_params : _type_, optional
        _description_, by default None
    data_params : _type_, optional
        _description_, by default None
    model_params : _type_, optional
        _description_, by default None
    train_params : _type_, optional
        _description_, by default None
    """
    tokenizer_params = tokenizer_params if tokenizer_params else {}
    data_params = data_params if data_params is not None else {}
    model_params = model_params if model_params is not None else {}
    train_params = train_params if train_params is not None else {}
    w2v_params = w2v_params if w2v_params is not None else {}
    default_data_formation = {
        "ques_content": "ques_content",
        "knowledge": "knowledge"
    }
    data_formation = data_params.get("data_formation", None)
    if data_formation is not None:
        default_data_formation.update(data_formation)
    data_formation = default_data_formation

    # tokenizer configuration
    if pretrained_dir is not None and os.path.exists(pretrained_dir):
        tokenizer = DisenQTokenizer.from_pretrained(pretrained_dir, **tokenizer_params)
    else:
        work_tokenizer_params = {
            "add_specials": None,
            "tokenize_method": "pure_text",
        }
        work_tokenizer_params.update(tokenizer_params if tokenizer_params else {})
        tokenizer = DisenQTokenizer(**work_tokenizer_params)
        corpus_items = train_items
        tokenizer.set_vocab(corpus_items,
                            key=lambda x: x[data_formation['ques_content']])

    # training Configuration
    work_train_params = deepcopy(DEFAULT_TRAIN_PARAMS)
    work_train_params["output_dir"] = output_dir
    if train_params is not None:
        work_train_params.update(train_params if train_params else {})
    if model_params:
        if 'hidden_size' in model_params:
            work_train_params['hidden_size'] = model_params['hidden_size']

    # dataset configuration
    items = train_items + ([] if eval_items is None else eval_items)
    if pretrained_dir:
        tokenizer, concept_to_idx, word2vec = preprocess_dataset(pretrained_dir, tokenizer, items,
                                                                 data_formation,
                                                                 trim_min_count=work_train_params["trim_min"],
                                                                 embed_dim=work_train_params["hidden_size"],
                                                                 w2v_params=w2v_params, silent=False)
    else:
        tokenizer, concept_to_idx, word2vec = preprocess_dataset(output_dir, tokenizer, items,
                                                                 data_formation,
                                                                 trim_min_count=work_train_params["trim_min"],
                                                                 embed_dim=work_train_params["hidden_size"],
                                                                 w2v_params=w2v_params, silent=False)
    train_dataset = DisenQDataset(train_items, tokenizer, data_formation,
                                  mode="train", concept_to_idx=concept_to_idx)
    if eval_items:
        eval_dataset = DisenQDataset(eval_items, tokenizer, data_formation,
                                     mode="test", concept_to_idx=concept_to_idx)
    else:
        eval_dataset = None

    # model configuration
    if pretrained_dir:
        model = DisenQNetForPreTraining.from_pretrained(pretrained_dir, **model_params)
    else:
        work_model_params = {
            "vocab_size": len(tokenizer),
            "hidden_size": 300,
            'concept_size': len(concept_to_idx),
            'dropout_rate': 0.2,
            'pos_weight': 1,
            'w_cp': 1.5,
            'w_mi': 1.0,
            'w_dis': 2.0,
            'warmup': 1,
            'n_adversarial': 10,
            'gamma': 0.5,
            'step_size': 20,
            'wv': word2vec
        }
        work_model_params.update(model_params if model_params else {})
        model = DisenQNetForPreTraining(**work_model_params)

    # Train
    work_args = DisenQTrainingArguments(**work_train_params)
    trainer = DisenQTrainer(
        model=model,
        args=work_args,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        data_collator=train_dataset.collate_fn,
    )
    trainer.train()
    # trainer.model.save_pretrained(output_dir)
    assert isinstance(trainer.model, DisenQNetForPreTraining)
    trainer.save_model(output_dir)
    trainer.model.save_config(output_dir)
    tokenizer.save_pretrained(output_dir)
    return output_dir



# coding: utf-8
# 2021/8/1 @ tongshiwei

from pickle import NONE
from typing import Iterable, Union
from ..SIF.segment import seg
from ..SIF.tokenization import tokenize
from ..SIF.tokenization.text import tokenize as tokenize_text
from ..SIF import sif4sci


__all__ = ["TOKENIZER", "Tokenizer", "CustomTokenizer", "CharTokenizer", "SpaceTokenizer",
           "PureTextTokenizer", "AstFormulaTokenizer", "get_tokenizer"]


class Tokenizer(object):
    """Iterator genetator for tokenization"""
    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def _tokenize(self, *args, **kwargs):
        raise NotImplementedError


class CustomTokenizer(Tokenizer):
    def __init__(self, symbol="gmas", figures=None, **kwargs):
        """Tokenize SIF items by customized configuration

        Parameters
        ----------
        symbol : str, optional
            Elements to symbolize before tokenization, by default "gmas"
        figures : _type_, optional
            Info for figures in items, by default None
        kwargs: addtional configuration for SIF items
            including text_params, formula_params, figure_params, more details could be found in `EduNLP.SIF.sif4sci`
        """
        self.tokenization_params = {
            "text_params": kwargs.get("text_params", None),
            "formula_params": kwargs.get("formula_params", None),
            "figure_params": kwargs.get("figure_params", None)
        }
        self.symbol = symbol
        self.figures = figures

    def __call__(self, items: Iterable, key=lambda x: x, **kwargs):
        """Tokenize items, return iterator genetator

        Parameters
        ----------
        item : Iterable
            question items
        key : function, optional
            determine how to get the text of items, by default lambdax: x
        """
        for item in items:
            yield self._tokenize(item, key=key, **kwargs)

    def _tokenize(self, item: Union[str, dict], key=lambda x: x, **kwargs):
        """Tokenize one item, return token list

        Parameters
        ----------
        item : Union[str, dict]
            question item
        key : function, optional
            determine how to get the text of item, by default lambdax: x
        """
        return tokenize(seg(key(item), symbol=self.symbol, figures=self.figures),
                        **self.tokenization_params, **kwargs).tokens


class CharTokenizer(Tokenizer):
    def __init__(self, stop_words="punctuations", **kwargs) -> None:
        """Tokenize text char by char. eg. "题目内容" -> ["题",  "目",  "内", 容"]

        Parameters
        ----------
        stop_words : str, optional
            stop_words to skip, by default "default"
        """
        self.stop_words = set("\n\r\t .,;?\"\'。．，、；？“”‘’（）") if stop_words == "punctuations" else stop_words
        self.stop_words = stop_words if stop_words is not None else set()

    def __call__(self, items: Iterable, key=lambda x: x, **kwargs):
        for item in items:
            yield self._tokenize(item, key=key, **kwargs)

    def _tokenize(self, item: Union[str, dict], key=lambda x: x, **kwargs):
        tokens = tokenize_text(key(item).strip(), granularity="char", stopwords=self.stop_words)
        return tokens


class SpaceTokenizer(Tokenizer):
    """Tokenize text by space. eg. "题目 内容" -> ["题目", "内容"]

    Parameters
    ----------
    stop_words : str, optional
        stop_words to skip, by default "default"
    """
    def __init__(self, stop_words="punctuations", **kwargs) -> None:
        stop_words = set("\n\r\t .,;?\"\'。．，、；？“”‘’（）") if stop_words == "punctuations" else stop_words
        self.stop_words = stop_words if stop_words is not None else set()

    def __call__(self, items: Iterable, key=lambda x: x, **kwargs):
        for item in items:
            yield self._tokenize(item, key=key, **kwargs)

    def _tokenize(self, item: Union[str, dict], key=lambda x: x, **kwargs):
        tokens = key(item).strip().split(' ')
        if self.stop_words:
            tokens = [w for w in tokens if w != '' and w not in self.stop_words]
        return tokens


class PureTextTokenizer(Tokenizer):
    def __init__(self, handle_figure_formula="skip", **kwargs):
        """
        Treat all elements in SIF item as prue text. Spectially, tokenize formulas as text.

        Parameters
        ----------
        handle_figure_formula : str, optional
            whether to skip or symbolize special formulas( $\\FormFigureID{…}$ and $\\FormFigureBase64{…}),
            by default skip

        Examples
        --------
        >>> tokenizer = PureTextTokenizer()
        >>> items = ["有公式$\\FormFigureID{1}$，如图$\\FigureID{088f15ea-xxx}$,\
        ... 若$x,y$满足约束条件公式$\\FormFigureBase64{2}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"]
        >>> tokens = tokenizer(items)
        >>> next(tokens)[:10]
        ['公式', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[SEP]', 'z']
        >>> items=["已知集合$A=\\left\\{x \\mid x^{2}-3 x-4<0\\right\\}, \\quad B=\\{-4,1,3,5\\}, \\quad$ 则 $A \\cap B=$"]
        >>> tokens = tokenizer(items)
        >>> next(tokens)  # doctest: +NORMALIZE_WHITESPACE
        ['已知', '集合', 'A', '=', '\\\\left', '\\\\{', 'x', '\\\\mid', 'x', '^', '{', '2', '}', '-', '3', 'x', '-', ...]
        >>> items = [{
        ... "stem": "已知集合$A=\\left\\{x \\mid x^{2}-3 x-4<0\\right\\}, \\quad B=\\{-4,1,3,5\\}, \\quad$ 则 $A \\cap B=$",
        ... "options": ["1", "2"]
        ... }]
        >>> tokens = tokenizer(items, key=lambda x: x["stem"])
        >>> next(tokens)  # doctest: +NORMALIZE_WHITESPACE
        ['已知', '集合', 'A', '=', '\\\\left', '\\\\{', 'x', '\\\\mid', 'x', '^', '{', '2', '}', '-', '3', 'x', '-', ...]
        >>> tokenizer = PureTextTokenizer(symbolize_figure_formula=True)
        >>> items = ["有公式$\\FormFigureID{1}$，如图$\\FigureID{088f15ea-xxx}$,\
        ... 若$x,y$满足约束条件公式$\\FormFigureBase64{2}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"]
        >>> tokens = tokenizer(items)
        >>> next(tokens)[:10]
        ['公式', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[SEP]', 'z']
        >>> items = ["$\\SIFTag{stem_begin}$若复数$z=1+2 i+i^{3}$，则$|z|=$$\\SIFTag{stem_end}$\
        ... $\\SIFTag{options_begin}$$\\SIFTag{list_0}$0$\\SIFTag{list_1}$1$\\SIFTag{list_2}$$\\sqrt{2}$\
        ... $\\SIFTag{list_3}$2$\\SIFTag{options_end}$"]
        >>> tokens = tokenizer(items)
        >>> next(tokens)[:10]
        ['[TAG]', '复数', 'z', '=', '1', '+', '2', 'i', '+', 'i']
        """
        # Formula images are skipped by default
        if handle_figure_formula == "skip":
            skip_figure_formula = True
            symbolize_figure_formula = False
        elif handle_figure_formula == "symbolize":
            skip_figure_formula = False
            symbolize_figure_formula = True
        elif handle_figure_formula is None:
            skip_figure_formula, symbolize_figure_formula = False, False
        else:
            raise ValueError('handle_figure_formula should be one in ["skip", "symbolize", None]')
        formula_params = {
            "method": "linear",
            "skip_figure_formula": skip_figure_formula,
            "symbolize_figure_formula": symbolize_figure_formula
        }
        text_params = {
            "granularity": "word",
            "stopwords": "default",
        }
        formula_params.update(kwargs.pop("formula_params", {}))
        text_params.update(kwargs.pop("text_params", {}))
        self.tokenization_params = {
            "formula_params": formula_params,
            "text_params": text_params,
            "figure_params": kwargs.get("figure_params", None)
        }

    def __call__(self, items: Iterable, key=lambda x: x, **kwargs):
        for item in items:
            yield self._tokenize(item, key=key, **kwargs)

    def _tokenize(self, item: Union[str, dict], key=lambda x: x, **kwargs):
        return tokenize(seg(key(item), symbol="gmas"), **self.tokenization_params, **kwargs).tokens


class AstFormulaTokenizer(Tokenizer):
    def __init__(self, symbol="gmas", figures=None, **kwargs):
        """Tokenize formulas in SIF items by AST parser.

        Parameters
        ----------
        symbol : str, optional
            Elements to symbolize before tokenization, by default "gmas"
        figures : _type_, optional
            Info for figures in items, by default None
        """
        formula_params = {
            "method": "ast",

            "ord2token": True,
            "return_type": "list",
            "var_numbering": True,

            "skip_figure_formula": False,
            "symbolize_figure_formula": True
        }
        text_params = {
            "granularity": "word",
            "stopwords": "default",
        }
        formula_params.update(kwargs.pop("formula_params", {}))
        text_params.update(kwargs.pop("text_params", {}))
        self.tokenization_params = {
            "formula_params": formula_params,
            "text_params": text_params,
            "figure_params": kwargs.pop("figure_params", None),
        }
        self.symbol = symbol
        self.figures = figures

    def __call__(self, items: Iterable, key=lambda x: x, **kwargs):
        for item in items:
            yield self._tokenize(item, key=key, **kwargs)

    def _tokenize(self, item: Union[str, dict], key=lambda x: x, **kwargs):
        mode = kwargs.pop("mode", 0)
        ret = sif4sci(key(item), figures=self.figures, symbol=self.symbol, mode=mode,
                      tokenization_params=self.tokenization_params, errors="ignore", **kwargs)
        ret = [] if ret is None else ret.tokens
        return ret


TOKENIZER = {
    "char": CharTokenizer,
    "space": SpaceTokenizer,
    "custom": CustomTokenizer,
    "pure_text": PureTextTokenizer,
    "ast_formula": AstFormulaTokenizer
}


def get_tokenizer(name, *args, **kwargs):
    r"""
    It is a total interface to use difference tokenizer.
    Parameters
    ----------
    name: str
        the name of tokenizer, e.g. text, pure_text.
    args:
        the parameters passed to tokenizer
    kwargs:
        the parameters passed to tokenizer
    Returns
    -------
    tokenizer: Tokenizer
    Examples
    --------
    >>> items = ["已知集合$A=\\left\\{x \\mid x^{2}-3 x-4<0\\right\\}, \\quad B=\\{-4,1,3,5\\}, \\quad$ 则 $A \\cap B=$"]
    >>> tokenizer = get_tokenizer("pure_text")
    >>> tokens = tokenizer(items)
    >>> next(tokens)  # doctest: +NORMALIZE_WHITESPACE
    ['已知', '集合', 'A', '=', '\\left', '\\{', 'x', '\\mid', 'x', '^', '{', '2', '}', '-', '3', 'x', '-', '4', '<',
    '0', '\\right', '\\}', ',', '\\quad', 'B', '=', '\\{', '-', '4', ',', '1', ',', '3', ',', '5', '\\}', ',',
    '\\quad', 'A', '\\cap', 'B', '=']
    """
    if name not in TOKENIZER:
        raise KeyError(
            "Unknown tokenizer %s, use one of the provided tokenizers: %s" % (name, ", ".join(TOKENIZER.keys()))
        )
    return TOKENIZER[name](*args, **kwargs)



from .base import Pipeline, PreProcessingPipeline
from .mappings import TASK_MAPPING, TOKENIZER_MAPPING_NAMES
from .property_prediction import PropertyPredictionPipeline
from .knowledge_prediction import KnowledgePredictionPipeline
from ..Pretrain import PretrainedEduTokenizer
from ..ModelZoo.base_model import BaseModel
from ..Vector.t2v import get_pretrained_model_info
from ..constant import MODEL_DIR
from EduData import get_data
from typing import Optional, Union, List

__all__ = ["pipeline"]

SUPPORTED_TASKS = {
    "pre-process": {
        "impl": Pipeline,
        "default": None
    },
    "property-prediction": {
        "impl": PropertyPredictionPipeline,
        "default": "elmo_for_property_prediction_test_256"
    },
    "knowledge-prediction": {
        "impl": KnowledgePredictionPipeline,
        "default": "elmo_for_knowledge_prediction_test_256"
    }
}


def pipeline(
        task: str = None,
        model: Optional[Union[BaseModel, str]] = None,
        tokenizer: Optional[PretrainedEduTokenizer] = None,
        pipeline_class: Optional[Pipeline] = None,
        preprocess: Optional[List] = None,
        **kwargs
):
    """
    Parameters
    ----------
    task: str, required
    model: BaseModel or str, optional

    tokenizer: PretrainedEduTokenizer, optional

    pipeline_class: Pipeline, optional
        to specify Pipeline class
    preprocess: list, optional
        a list of names of pre-process pipes

    Examples
    ----------
    >>> processor = pipeline(task="property-prediction") # doctest: +SKIP
    >>> item = "如图所示，则三角形ABC的面积是_。"
    >>> processor(item) # doctest: +SKIP
    """
    if preprocess is None and task is None and model is None:
        raise RuntimeError("Please specify at least the model to use or task to do!")
    elif model is None and tokenizer is not None:
        raise RuntimeError("Specified tokenizer but no model is not allowed!")
    elif task is None and model is not None:
        raise RuntimeError("Please specify the task.")
    elif task is None:
        task = "pre-process"

    if task == "pre-process":
        return PreProcessingPipeline(pipe_names=preprocess)

    if task in SUPPORTED_TASKS:
        targeted_task = SUPPORTED_TASKS[task]
    else:
        raise KeyError(f"Unknown task {task}")
    if pipeline_class is None:
        pipeline_class = targeted_task["impl"]
    if model is None or isinstance(model, str):
        # TODO: 1. waiting for ModelHub and TEST
        #       2. Check if the specified model and task are matched
        # pretrained_name = targeted_task["default"] if model is None else model
        # model_url, model_name, *args = get_pretrained_model_info(pretrained_name)
        # model_path = get_data(model_url, MODEL_DIR)
        # model = TASK_MAPPING[task][model_name].from_pretrained(model_path)
        # tokenizer = TOKENIZER_MAPPING_NAMES[model_name].from_pretrained(model_path)
        pass
    elif isinstance(model, BaseModel) and isinstance(tokenizer, PretrainedEduTokenizer):
        model, tokenizer = model, tokenizer
    elif model is not None and tokenizer is not None:
        raise KeyError(f"Unknown model and tokenizer: {model} and {tokenizer}")

    return pipeline_class(model=model, task=task, tokenizer=tokenizer, preproc_pipe_names=preprocess, **kwargs)



import torch


def sequence_mask(lengths, max_len=None):
    """Same as tf.sequence_mask, Returns a mask tensor representing the first N positions of each cell.

    Parameters
    ----------
    lengths : _type_
        integer tensor, all its values <= maxlen.
    max_len : _type_, optional
        scalar integer tensor, size of last dimension of returned tensor. Default is the maximum value in lengths.

    Returns
    -------
    _type_
        A mask tensor of shape lengths.shape + (maxlen,)

    Examples:
    ---------
    >>> sequence_mask(torch.tensor([1, 3, 2]), 5)
    tensor([[ True, False, False, False, False],
            [ True,  True,  True, False, False],
            [ True,  True, False, False, False]])
    >>> sequence_mask(torch.tensor([[1, 3],[2,0]]))
    tensor([[[ True, False, False],
             [ True,  True,  True]],
    <BLANKLINE>
            [[ True,  True, False],
             [False, False, False]]])
    """

    lengths_shape = lengths.shape  # torch.size() is a tuple
    lengths = lengths.reshape(-1)

    batch_size = lengths.numel()
    max_len = max_len or int(lengths.max())
    lengths_shape += (max_len,)

    return (torch.arange(0, max_len, device=lengths.device)
            .type_as(lengths)
            .unsqueeze(0).expand(batch_size, max_len)
            .lt(lengths.unsqueeze(1))).reshape(lengths_shape)


def gather_nd(params, indices):
    """_summary_

    Parameters
    ----------
    params : _type_
        _description_
    indices : _type_
        _description_

    Returns
    -------
    _type_
        _description_

    Examples:
    ---------
    >>> gather_nd(
    ...           params=torch.tensor([[1, 2, 3],
    ...                                [4, 5, 6]]),
    ...           indices=torch.tensor([[1],
    ...                                 [0]]))
    tensor([[4, 5, 6],
            [1, 2, 3]])
    """
    newshape = indices.shape[:-1] + params.shape[indices.shape[-1]:]
    indices = indices.view(-1, indices.shape[-1]).tolist()
    out = torch.cat([params.__getitem__(tuple(i)) for i in indices])
    return out.reshape(newshape)



# coding: utf-8
# 2021/5/29 @ tongshiwei

from .gensim_vec import W2V, D2V, BowLoader, TfidfLoader
from .const import *
from .rnn import RNNModel
from .t2v import T2V, get_pretrained_t2v, get_pretrained_model_info, get_all_pretrained_models
from .embedding import Embedding
from .bert_vec import BertModel
from .quesnet import QuesNetModel
from .disenqnet import DisenQModel
from .elmo_vec import ElmoModel



import torch
import torch.nn as nn
from torch.autograd import Variable
from transformers.modeling_outputs import ModelOutput
from typing import List, Dict, Optional

__all__ = ["HAM"]


class HAMOutput(ModelOutput):
    """
        Output type of [`HAM`]

        Parameters
        ----------
        scores: of shape (batch_size, sequence_length)
        first_att_weight: of shape
        second_visual: of shape (batch_size, num_classes)
        third_visual: of shape (batch_size, num_classes)
        first_logits: of shape (batch_size, num_classes)
        second_logits: of shape (batch_size, num_classes)
        third_logits: of shape (batch_size, num_classes)
        global_logits: of shape (batch_size, num_classes)
        first_scores: of shape (batch_size, num_classes)
        second_scores: of shape (batch_size, num_classes)
    """
    scores: torch.FloatTensor = None
    first_att_weight: torch.FloatTensor = None
    second_visual: torch.FloatTensor = None
    third_visual: torch.FloatTensor = None
    first_logits: torch.FloatTensor = None
    second_logits: torch.FloatTensor = None
    third_logits: torch.FloatTensor = None
    global_logits: torch.FloatTensor = None
    first_scores: torch.FloatTensor = None
    second_scores: torch.FloatTensor = None


def truncated_normal_(tensor, mean=0, std=0.1):
    size = tensor.shape
    tmp = tensor.new_empty(size + (4,)).normal_()
    valid = (tmp < 2) & (tmp > -2)
    ind = valid.max(-1, keepdim=True)[1]
    tensor.data.copy_(tmp.gather(-1, ind).squeeze(-1))
    tensor.data.mul_(std).add_(mean)
    return tensor


class AttentionLayer(nn.Module):
    def __init__(self, num_units, attention_unit_size, num_classes):
        super(AttentionLayer, self).__init__()
        self.fc1 = nn.Linear(num_units, attention_unit_size, bias=False)
        self.fc2 = nn.Linear(attention_unit_size, num_classes, bias=False)

    def forward(self, input_x):
        attention_matrix = self.fc2(torch.tanh(self.fc1(input_x))).transpose(1, 2)
        attention_weight = torch.softmax(attention_matrix, dim=-1)
        attention_out = torch.matmul(attention_weight, input_x)
        return attention_weight, torch.mean(attention_out, dim=1)


class LocalLayer(nn.Module):
    def __init__(self, num_units, num_classes):
        super(LocalLayer, self).__init__()
        self.fc = nn.Linear(num_units, num_classes)

    def forward(self, input_x, input_att_weight):
        logits = self.fc(input_x)
        scores = torch.sigmoid(logits)
        visual = torch.mul(input_att_weight, scores.unsqueeze(-1))
        visual = torch.softmax(visual, dim=-1)
        visual = torch.mean(visual, dim=1)
        return logits, scores, visual


class HAM(nn.Module):

    def __init__(
            self,
            num_classes_list: List[int],
            num_total_classes: int,
            sequence_model_hidden_size: int,
            attention_unit_size: Optional[int] = 256,
            fc_hidden_size: Optional[int] = 512,
            beta: Optional[float] = 0.5,
            dropout_rate=None):
        super(HAM, self).__init__()
        self.beta = beta

        assert num_total_classes == sum(num_classes_list)
        # First Level
        self.first_attention = AttentionLayer(sequence_model_hidden_size, attention_unit_size, num_classes_list[0])
        self.first_fc = nn.Linear(sequence_model_hidden_size * 2, fc_hidden_size)
        self.first_local = LocalLayer(fc_hidden_size, num_classes_list[0])

        # Second Level
        self.second_attention = AttentionLayer(sequence_model_hidden_size, attention_unit_size, num_classes_list[1])
        self.second_fc = nn.Linear(sequence_model_hidden_size * 2, fc_hidden_size)
        self.second_local = LocalLayer(fc_hidden_size, num_classes_list[1])

        # Third Level
        self.third_attention = AttentionLayer(sequence_model_hidden_size, attention_unit_size, num_classes_list[2])
        self.third_fc = nn.Linear(sequence_model_hidden_size * 2, fc_hidden_size)
        self.third_local = LocalLayer(fc_hidden_size, num_classes_list[2])

        # Fully Connected Layer
        self.fc = nn.Linear(fc_hidden_size * 3, fc_hidden_size)

        # Highway Layer
        self.highway_lin = nn.Linear(fc_hidden_size, fc_hidden_size)
        self.highway_gate = nn.Linear(fc_hidden_size, fc_hidden_size)

        # Add dropout
        self.dropout = nn.Dropout(dropout_rate)

        # Global scores
        self.global_scores_fc = nn.Linear(fc_hidden_size, num_total_classes)

        for name, param in self.named_parameters():
            if 'embedding' not in name and 'weight' in name:
                truncated_normal_(param.data, mean=0, std=0.1)
            else:
                nn.init.constant_(param.data, 0.1)

    def forward(self, sequential_embeddings):

        sequential_embeddings_pool = torch.mean(sequential_embeddings, dim=1)

        # First Level
        first_att_weight, first_att_out = self.first_attention(sequential_embeddings)
        first_local_input = torch.cat((sequential_embeddings_pool, first_att_out), dim=1)
        first_local_fc_out = self.first_fc(first_local_input)
        first_logits, first_scores, first_visual = self.first_local(first_local_fc_out, first_att_weight)

        # Second Level
        second_att_input = torch.mul(sequential_embeddings, first_visual.unsqueeze(-1))
        second_att_weight, second_att_out = self.second_attention(second_att_input)
        second_local_input = torch.cat((sequential_embeddings_pool, second_att_out), dim=1)
        second_local_fc_out = self.second_fc(second_local_input)
        second_logits, second_scores, second_visual = self.second_local(second_local_fc_out, second_att_weight)

        # Third Level
        third_att_input = torch.mul(sequential_embeddings, second_visual.unsqueeze(-1))
        third_att_weight, third_att_out = self.third_attention(third_att_input)
        third_local_input = torch.cat((sequential_embeddings_pool, third_att_out), dim=1)
        third_local_fc_out = self.third_fc(third_local_input)
        third_logits, third_scores, third_visual = self.third_local(third_local_fc_out, third_att_weight)

        # Concat
        # shape of ham_out: [batch_size, fc_hidden_size * 3]
        ham_out = torch.cat((first_local_fc_out, second_local_fc_out,
                             third_local_fc_out), dim=1)

        # Fully Connected Layer
        fc_out = self.fc(ham_out)

        # Highway Layer and Dropout
        highway_g = torch.relu(self.highway_lin(fc_out))
        highway_t = torch.sigmoid(self.highway_gate(fc_out))
        highway_output = torch.mul(highway_g, highway_t) + torch.mul((1 - highway_t), fc_out)
        h_drop = self.dropout(highway_output)

        # Global scores
        global_logits = self.global_scores_fc(h_drop)
        global_scores = torch.sigmoid(global_logits)
        local_scores = torch.cat((first_scores, second_scores, third_scores), dim=1)
        scores = self.beta * global_scores + (1 - self.beta) * local_scores
        return HAMOutput(
            scores=scores,
            first_att_weight=first_att_weight,
            first_visual=first_visual,
            second_visual=second_visual,
            third_visual=third_visual,
            first_logits=first_logits,
            second_logits=second_logits,
            third_logits=third_logits,
            global_logits=global_logits,
            first_scores=first_scores,
            second_scores=second_scores
        )



import torch
from torch import nn
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence
from baize.torch import load_net
import torch.nn.functional as F
import json
import os
from ..base_model import BaseModel
from transformers.modeling_outputs import ModelOutput
from transformers import BertModel, PretrainedConfig
from typing import List, Optional
from ..rnn.harnn import HAM

__all__ = ["BertForPropertyPrediction", "BertForKnowledgePrediction"]


class BertForPPOutput(ModelOutput):
    loss: torch.FloatTensor = None
    logits: torch.FloatTensor = None


class BertForPropertyPrediction(BaseModel):
    def __init__(self, pretrained_model_dir=None, head_dropout=0.5):
        super(BertForPropertyPrediction, self).__init__()
        self.bert = BertModel.from_pretrained(pretrained_model_dir)
        self.hidden_size = self.bert.config.hidden_size
        self.head_dropout = head_dropout
        self.dropout = nn.Dropout(head_dropout)
        self.classifier = nn.Linear(self.hidden_size, 1)
        self.sigmoid = nn.Sigmoid()
        self.criterion = nn.MSELoss()

        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__"]}
        self.config['architecture'] = 'BertForPropertyPrediction'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self,
                input_ids=None,
                attention_mask=None,
                token_type_ids=None,
                labels=None):
        outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        item_embeds = outputs.last_hidden_state[:, 0, :]
        item_embeds = self.dropout(item_embeds)

        logits = self.sigmoid(self.classifier(item_embeds)).squeeze(1)
        loss = None
        if labels is not None:
            loss = self.criterion(logits, labels) if labels is not None else None
        return BertForPPOutput(
            loss=loss,
            logits=logits,
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                pretrained_model_dir=model_config['pretrained_model_dir'],
                head_dropout=model_config.get("head_dropout", 0.5)
            )

    # @classmethod
    # def from_pretrained(cls):
    #     NotImplementedError
    #     # 需要验证是否和huggingface的模型兼容


class BertForKnowledgePrediction(BaseModel):
    def __init__(self,
                 num_classes_list: List[int] = None,
                 num_total_classes: int = None,
                 pretrained_model_dir=None,
                 head_dropout=0.5,
                 flat_cls_weight=0.5,
                 attention_unit_size=256,
                 fc_hidden_size=512,
                 beta=0.5,
                 ):
        super(BertForKnowledgePrediction, self).__init__()
        self.bert = BertModel.from_pretrained(pretrained_model_dir)
        self.hidden_size = self.bert.config.hidden_size
        self.head_dropout = head_dropout
        self.dropout = nn.Dropout(head_dropout)
        self.classifier = nn.Linear(self.hidden_size, 1)
        self.sigmoid = nn.Sigmoid()
        self.criterion = nn.MSELoss()
        self.flat_classifier = nn.Linear(self.hidden_size, num_total_classes)
        self.ham_classifier = HAM(
            num_classes_list=num_classes_list,
            num_total_classes=num_total_classes,
            sequence_model_hidden_size=self.bert.config.hidden_size,
            attention_unit_size=attention_unit_size,
            fc_hidden_size=fc_hidden_size,
            beta=beta,
            dropout_rate=head_dropout
        )
        self.flat_cls_weight = flat_cls_weight
        self.num_classes_list = num_classes_list
        self.num_total_classes = num_total_classes

        self.config = {k: v for k, v in locals().items() if k not in ["self", "__class__"]}
        self.config['architecture'] = 'BertForKnowledgePrediction'
        self.config = PretrainedConfig.from_dict(self.config)

    def forward(self,
                input_ids=None,
                attention_mask=None,
                token_type_ids=None,
                labels=None):
        outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        item_embeds = outputs.last_hidden_state[:, 0, :]
        item_embeds = self.dropout(item_embeds)
        tokens_embeds = outputs.last_hidden_state
        tokens_embeds = self.dropout(tokens_embeds)
        flat_logits = self.sigmoid(self.flat_classifier(item_embeds))
        ham_outputs = self.ham_classifier(tokens_embeds)
        ham_logits = self.sigmoid(ham_outputs.scores)
        logits = self.flat_cls_weight * flat_logits + (1 - self.flat_cls_weight) * ham_logits
        loss = None
        if labels is not None:
            labels = torch.sum(torch.nn.functional.one_hot(labels, num_classes=self.num_total_classes), dim=1)
            labels = labels.float()
            loss = self.criterion(logits, labels) if labels is not None else None
        return BertForPPOutput(
            loss=loss,
            logits=logits,
        )

    @classmethod
    def from_config(cls, config_path, **kwargs):
        with open(config_path, "r", encoding="utf-8") as rf:
            model_config = json.load(rf)
            model_config.update(kwargs)
            return cls(
                pretrained_model_dir=model_config['pretrained_model_dir'],
                head_dropout=model_config.get("head_dropout", 0.5),
                num_classes_list=model_config.get('num_classes_list'),
                num_total_classes=model_config.get('num_total_classes'),
                flat_cls_weight=model_config.get('flat_cls_weight', 0.5),
                attention_unit_size=model_config.get('attention_unit_size', 256),
                fc_hidden_size=model_config.get('fc_hidden_size', 512),
                beta=model_config.get('beta', 0.5),
            )

    # @classmethod
    # def from_pretrained(cls):
    #     NotImplementedError
    #     # 需要验证是否和huggingface的模型兼容



# -*- coding: utf-8 -*-

import torch
from torch import nn
from torch.nn import functional as F

from .utils import get_mask, shuffle, spectral_norm
from ..utils import MLP, TextCNN


class TextEncoder(nn.Module):
    """
    Parameters
    ----------
    vocab_size: int
        size of vocabulary
    hidden_dim: int
        size of word and question embedding
    dropout: float
        dropout rate
    wv: torch.Tensor
        Tensor of (vocab_size, hidden_dim), initial word embedding, default = None
    """
    def __init__(self, vocab_size, hidden_dim, dropout, wv=None):
        super(TextEncoder, self).__init__()
        self.embed = nn.Embedding(vocab_size, hidden_dim)
        if wv is not None:
            self.load_wv(wv)
        self.encoder = TextCNN(hidden_dim, hidden_dim)
        self.dropout = nn.Dropout(p=dropout)
        return

    def forward(self, input):
        """
        Parameters
        ----------
        input:
            Tensor of long (batch_size, seq_len), word index

        Returns
        -------
        (embed, hidden)
            - embed: Tensor of (batch_size, seq_len, hidden_dim), word embedding
            - hidden: Tensor of (batch_size, hidden_dim), question embedding
        """
        embed = self.embed(input)
        embed_dp = self.dropout(embed)
        hidden = self.encoder(embed_dp)
        return embed, hidden

    def load_wv(self, wv: torch.Tensor):
        self.embed.weight.data.copy_(wv)


class AttnModel(nn.Module):
    """
    Attention Model

    Parameters
    ----------
    dim: int
        size of hidden vector
    dropout: float
        dropout rate of attention model
    """

    def __init__(self, dim, dropout):
        super(AttnModel, self).__init__()
        self.score = MLP(dim * 2, 1, dim, 0, n_layers=2, act=torch.tanh)
        self.fw = nn.Linear(dim, dim)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, q, k, v, mask=None):
        """
        Parameters
        ----------
        q: torch.Tensor
            Tensor of (batch_size, hidden_size) or (batch_size, out_size, hidden_size)
            query, out_size = 1 if discarded
        k: torch.Tensor
            Tensor of (batch_size, in_size, hidden_size), key
        v: torch.Tensor
            Tensor of (batch_size, in_size, hidden_size), value
        mask: torch.Tensor
            Tensor of (batch_size, in_size),
            key/value mask, where 0 means data and 1 means pad, default = None, means zero matrix

        Returns
        -------
        (output, attn)
            - output: Tensor of (batch_size, hidden_size) or (batch_size, out_size, hidden_size),
            attention output, shape according to q
            - attn: Tensor of (batch_size, in_size) or (batch_size, out_size, in_size),
            attention weight, shape according to q
        """
        # (b, h) -> (b, o, h)
        q_dim = q.dim()
        if q_dim == 2:
            q = q.unsqueeze(1)
        # (b, o, i, h)
        output_size = q.size(1)
        input_size = k.size(1)
        qe = q.unsqueeze(2).expand(-1, -1, input_size, -1)
        ke = k.unsqueeze(1).expand(-1, output_size, -1, -1)
        # (b, o, i, h) -> (b, o, i, 1) -> (b, o, i)
        score = self.score(torch.cat((qe, ke), dim=-1)).squeeze(-1)
        if mask is not None:
            score.masked_fill_(mask.unsqueeze(1).expand(-1, output_size, -1), -float('inf'))
        attn = F.softmax(score, dim=-1)
        # (b, o, i) * (b, i, h) -> (b, o, h)
        output = torch.bmm(attn, v)
        # (b, o, h) -> (b, h), (b, o, i) -> (b, i)
        if q_dim == 2:
            output = output.squeeze(1)
            attn = attn.squeeze(1)
        output = F.leaky_relu(self.fw(self.dropout(output)))
        return output, attn


class ConceptEstimator(nn.Module):
    """
    Concept estimator
    min Classifier_loss

    Parameters
    ----------
    hidden_dim: int
        size of question embedding
    concept_size: int
        number of concept classes
    pos_weight: float
        positive sample weight in unbalanced multi-label concept classifier
    dropout: float
        dropout rate
    """

    def __init__(self, hidden_dim, concept_size, pos_weight, dropout):
        super(ConceptEstimator, self).__init__()
        self.classifier = MLP(hidden_dim, concept_size, hidden_dim, dropout, n_layers=2)
        pos_weight = torch.ones(concept_size) * pos_weight
        self.loss = nn.BCEWithLogitsLoss(reduction="mean", pos_weight=pos_weight)
        return

    def forward(self, hidden, label):
        """
        Parameters
        ----------
        hidden: torch.Tensor
            Tensor of (batch_size, hidden_dim), question embedding
        label: torch.Tensor
            Tensor of long (batch_size, concept_size), question concept one-hot label

        Returns
        -------
        loss: float
            Tensor of (), BCE loss of concept label
        """
        output = self.classifier(hidden)
        loss = self.loss(output, label.float())
        return loss


class MIEstimator(nn.Module):
    """
    MI estimator between question and word embedding for maximization
    max MI(X,Y)

    Parameters
    ----------
    embed_dim: int
        size of word embedding
    hidden_dim: int
        size of question embedding
    dropout: float
        dropout rate
    """

    def __init__(self, embed_dim, hidden_dim, dropout):
        super(MIEstimator, self).__init__()
        self.disc = Disc(embed_dim, hidden_dim, dropout)
        return

    def shuffle_in_batch(self, x):
        # |0 0 0 0 0|    |1 2 3 1 2|
        # |1 1 1 1 1| => |2 3 0 2 3|
        # |2 2 2 2 2|    |3 0 1 3 0|
        # |3 3 3 3 3|    |0 1 2 0 1|
        device = x.device
        batch_size, seq_len, _ = x.size()
        if batch_size == 1:
            sx = x
        else:
            batch_index = torch.arange(batch_size).unsqueeze(1)
            shuffle_index = (torch.arange(batch_size - 1) + 1)
            shuffle_index = shuffle_index.repeat(seq_len // (batch_size - 1) + 1)[:seq_len].unsqueeze(0)
            batch_index = batch_index.to(device)
            shuffle_index = shuffle_index.to(device)
            shuffled_batch_index = (batch_index + shuffle_index) % batch_size
            seq_batch_index = torch.arange(seq_len).unsqueeze(0).expand(batch_size, -1)
            seq_batch_index = seq_batch_index.to(device)
            sx = x[shuffled_batch_index, seq_batch_index]
        return sx

    def forward(self, embed, hidden, lengths):
        """
        average MI between hidden and each word embedding
        MI(X,Y) = Avg(E_pxy[-sp(-T(xi,y))] - E_pxpy[sp(T(xi,y))])

        Parameters
        ----------
        embed: torch.Tensor
            Tensor of (batch_size, seq_len, embed_dim), word embedding vector
        hidden: torch.Tensor
            Tensor of (batch_size, hidden_dim), question hidden vector
        lengths: torch.Tensor
            Tensor of (batch_size), length of valid words in each question

        Returns
        -------
        mi: float
            mi between question and each word embedding
        """
        # batch_size * seq_len * hidden_dim
        seq_len = embed.size(1)
        hidden = hidden.unsqueeze(1).expand(-1, seq_len, -1)
        s_hidden = self.shuffle_in_batch(hidden)
        # batch_size * seq_len, [F F F T T T], False means data, True means pad
        mask = get_mask(seq_len, lengths)
        # seq_len
        seq_batch_len = (~mask).sum(dim=0)
        # P(X,Y) = (x, y), P(X)P(Y) = (x, sy)
        # batch_size * seq_len -> seq_len
        # -sp(disc(x, y)).mean() - sp(disc(x, sy)).mean()
        seq_mi = - F.softplus(-self.disc(embed, hidden)).masked_fill(mask, 0).sum(dim=0) / seq_batch_len\
                 - F.softplus(self.disc(embed, s_hidden)).masked_fill(mask, 0).sum(dim=0) / seq_batch_len
        # seq_len -> 1
        mi = seq_mi.mean()
        return mi


class DisenEstimator(nn.Module):
    """
    Disentangling estimator by WGAN-like adversarial training and spectral normalization for MI minimization
    MI(X,Y) = E_pxy[T(x,y)] - E_pxpy[T(x,y)]
    min_xy max_T MI(X,Y)

    Parameters
    ----------
    hidden_dim: int
        size of question embedding
    dropout: float
        dropout rate
    """

    def __init__(self, hidden_dim, dropout):
        super(DisenEstimator, self).__init__()
        self.disc = Disc(hidden_dim, hidden_dim, dropout)
        return

    def forward(self, x, y):
        """
        Parameters
        ----------
        x: torch.Tensor
            Tensor of (batch_size, hidden_dim), x
        y: torch.Tensor
            Tensor of (batch_size, hidden_dim), y

        Returns
        -------
        loss:
            loss for MI minimization
        """
        sy = shuffle(y)
        loss = self.disc(x, y).mean() - self.disc(x, sy).mean()
        return loss

    def spectral_norm(self):
        """
            spectral normalization to satisfy Lipschitz constrain for Disc of WGAN
        """
        # Lipschitz constrain for Disc of WGAN
        with torch.no_grad():
            for w in self.parameters():
                w.data /= spectral_norm(w.data)
        return


class Disc(nn.Module):
    def __init__(self, x_dim, y_dim, dropout):
        super(Disc, self).__init__()
        self.disc = MLP(x_dim + y_dim, 1, y_dim, dropout, n_layers=2)

    def forward(self, x, y):
        input = torch.cat((x, y), dim=-1)
        # (b, 1) -> (b)
        score = self.disc(input).squeeze(-1)
        return score



# coding: utf-8
# 2021/7/13 @ tongshiwei

import os
import json
import requests
from longling import path_append
from EduData import get_data
from .rnn import RNNModel
from .gensim_vec import W2V, D2V
from .bert_vec import BertModel
from .quesnet import QuesNetModel
from .elmo_vec import ElmoModel
from .meta import Vector
from EduNLP.constant import MODEL_DIR
from .disenqnet import DisenQModel


MODELS = {
    "w2v": W2V,
    "d2v": D2V,
    "rnn": RNNModel,
    "lstm": RNNModel,
    "gru": RNNModel,
    "elmo": ElmoModel,
    'bert': BertModel,
    'quesnet': QuesNetModel,
    "disenq": DisenQModel,
}


MODELHUB_URL = 'https://modelhub-backend-269-production.env.bdaa.pro/v1/api/'


class T2V(object):
    """
    The function aims to transfer token list to vector. If you have a certain model, you can use T2V directly. \
    Otherwise, calling get_pretrained_t2v function is a better way to get vector which can switch it without your model.

    Parameters
    ----------
    model: str
        select the model type
        e.g.: d2v, rnn, lstm, gru, elmo, etc.

    Examples
    --------
    >>> item = [{'ques_content':'有公式$\\FormFigureID{wrong1?}$和公式$\\FormFigureBase64{wrong2?}$，\
    ... 如图$\\FigureID{088f15ea-8b7c-11eb-897e-b46bfc50aa29}$,若$x,y$满足约束条件$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$'}]
    >>> model_dir = "examples/test_model/d2v"
    >>> url, model_name, *args = get_pretrained_model_info('d2v_test_256')
    >>> (); path = get_data(url, model_dir); () # doctest: +ELLIPSIS
    (...)
    >>> path = path_append(path, os.path.basename(path) + '.bin', to_str=True)
    >>> t2v = T2V('d2v',filepath=path)
    >>> print(t2v(item))
    [array([...dtype=float32)]
    """

    def __init__(self, model: str, *args, **kwargs):
        model = model.lower()
        self.model_type = model
        if model in {"rnn", "lstm", "gru"}:
            self.i2v: Vector = MODELS[model](model, *args, **kwargs)
        else:
            self.i2v: Vector = MODELS[model](*args, **kwargs)

    def __call__(self, items, *args, **kwargs):
        return self.i2v.infer_vector(items, *args, **kwargs)

    def infer_vector(self, items, *args, **kwargs):
        return self.i2v.infer_vector(items, *args, **kwargs)

    def infer_tokens(self, items, *args, **kwargs):
        return self.i2v.infer_tokens(items, *args, **kwargs)

    @property
    def vector_size(self) -> int:
        return self.i2v.vector_size


def get_pretrained_model_info(name):
    url = MODELHUB_URL + 'getPretrainedModel'
    param = {'name': name}
    r = requests.get(url, params=param)
    assert r.status_code == 200, r.status_code
    r = json.loads(r.content)
    return [r['url'], r['t2v_name']]


def get_all_pretrained_models():
    url = MODELHUB_URL + 'getPretrainedModelList'
    r = requests.get(url)
    assert r.status_code == 200, r.status_code
    r = json.loads(r.content)
    return r['name']


def get_pretrained_t2v(name, model_dir=MODEL_DIR, **kwargs):
    """
    It is a good idea if you want to switch token list to vector earily.

    Parameters
    ----------
    name:str
        select the pretrained model
        e.g.:
        d2v_math_300
        w2v_math_300
        elmo_math_2048
        bert_math_768
        bert_taledu_768
        disenq_math_256
        quesnet_math_512
    model_dir:str
        the path of model, default: MODEL_DIR = '~/.EduNLP/model'

    Returns
    -------
    t2v model: T2V

    Examples
    --------
    >>> item = [{'ques_content':'有公式$\\FormFigureID{wrong1?}$和公式$\\FormFigureBase64{wrong2?}$，\
    ... 如图$\\FigureID{088f15ea-8b7c-11eb-897e-b46bfc50aa29}$,若$x,y$满足约束条件$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$'}]
    >>> i2v = get_pretrained_t2v("d2v_test_256", "examples/test_model/d2v") # doctest: +SKIP
    >>> print(i2v(item)) # doctest: +SKIP
    [array([...dtype=float32)]
    """
    pretrained_models = get_all_pretrained_models()
    if name not in pretrained_models:
        raise KeyError(
            "Unknown pretrained model %s, use one of the provided pretrained models: %s" % (
                name, ", ".join(pretrained_models))
        )
    url, model_name, *args = get_pretrained_model_info(name)
    model_path = get_data(url, model_dir)
    if model_name in ["d2v", "w2v"]:
        postfix = ".bin" if model_name == "d2v" else ".kv"
        model_path = path_append(model_path, os.path.basename(model_path) + postfix, to_str=True)
    return T2V(model_name, model_path, *args, **kwargs)



# coding: utf-8
# 2021/5/16 @ tongshiwei

import traceback
import warnings
from .segment import seg
from .tokenization import tokenize, link_formulas
from .parser import Parser

__all__ = ["is_sif", "to_sif", "sif4sci"]


def is_sif(item, check_formula=True, return_parser=False):
    r"""
    the part aims to check whether the input is sif format

    Parameters
    ----------
    item:str
        a raw item which respects stem
    check_formula: bool
        whether to check the formulas when parsing item.

        True if check the validity of formulas in item
        False if not check the validity of formulas in item, which is faster
    return_parser: bool
        whether to put the parsed item in return.

        when True, the format of return is (bool, Parser)
        when False, the format of return is bool

    Returns
    -------
    bool
        when item can not be parsed correctly, raise ValueError;
        when item is in stardarded format originally, return Ture (and the Parser of item);
        when item isn't in stardarded format originally, return False (and the Parser of item);

    Examples
    --------
    >>> text = '若$x,y$满足约束条件' \
    ...        '$\\left\\{\\begin{array}{c}2 x+y-2 \\leq 0 \\\\ x-y-1 \\geq 0 \\\\ y+1 \\geq 0\\end{array}\\right.$，' \
    ...        '则$z=x+7 y$的最大值$\\SIFUnderline$'
    >>> is_sif(text)
    True
    >>> text = '某校一个课外学习小组为研究某作物的发芽率y和温度x（单位...'
    >>> ret = is_sif(text, return_parser=True)
    >>> ret # doctest: +ELLIPSIS
    (False, <EduNLP.SIF.parser.parser.Parser object...>)
    """
    item_parser = Parser(item, check_formula)
    item_parser.description_list()
    if item_parser.fomula_illegal_flag:
        raise ValueError(item_parser.fomula_illegal_message)
    ret = True if item_parser.error_flag == 0 and item_parser.modify_flag == 0 else False
    if return_parser is True:
        return ret, item_parser
    else:
        return ret


def to_sif(item, check_formula=True, parser: Parser = None):
    r"""
    the part aims to switch item to sif formate

    Parameters
    ----------
    items:str
        a raw item which respects stem
    check_formula: bool
        whether to check the formulas when parsing item (only work when parser=None).
    parser: Parser
        the parser of item returned from is_sif.

    Returns
    -------
    item:str
        the item which accords with sif format

    Examples
    --------
    >>> text = '某校一个课外学习小组为研究某作物的发芽率y和温度x（单位...'
    >>> siftext = to_sif(text)
    >>> siftext
    '某校一个课外学习小组为研究某作物的发芽率$y$和温度$x$（单位...'
    >>> ret = is_sif(text, return_parser=True)
    >>> ret # doctest: +ELLIPSIS
    (False, <EduNLP.SIF.parser.parser.Parser object...>)
    >>> to_sif(text, parser=ret[1])
    '某校一个课外学习小组为研究某作物的发芽率$y$和温度$x$（单位...

    """
    if parser is not None:
        return parser.text
    else:
        return is_sif(item, check_formula, return_parser=True)[1].text


def sif4sci(item: str, figures: (dict, bool) = None, mode: int = 2, symbol: str = None, tokenization=True,
            tokenization_params=None, errors="raise"):
    r"""

    Default to use linear Tokenizer, change the tokenizer by specifying tokenization_params

    Parameters
    ----------
    item:str
        a raw item which respects stem
    figures:dict
        when it is a dict, it means the id-to-instance for figures in 'FormFigureID{...}' format,
        when it is a bool, it means whether to instantiate figures in 'FormFigureBase64{...}' format

    mode: int
        when safe = 2, use is_sif and check formula in item
        when safe = 1, use is_sif but don't check formula in item
        when safe = 0, don't use is_sif and don't check anything in item

    symbol: str
        select the methods to symbolize:
            "t": text
            "f": formula
            "g": figure
            "m": question mark
            "a": tag
            "s": sep

    tokenization: bool
        whether to tokenize item after segmentation

    tokenization_params:
        the dict of text_params, formula_params and figure_params in tokenization
        For formula_params:
            method: which tokenizer to be used, "linear" or "ast"
            The parameters only useful for "linear":
                skip_figure_formula: whether to skip the formula in figure format
                symbolize_figure_formula: whether to symbolize the formula in figure format
            The parameters only useful for "ast":
                ord2token: whether to transfer the variables (mathord) and constants (textord) to special tokens.
                var_numbering: whether to use number suffix to denote different variables
                return_type: 'list' or 'ast'
            More parameters can be found in the definition in SIF.tokenization.formula
        For figure_params:
            figure_instance：whether to return instance of figures in tokens
        For text_params: See definition in SIF.tokenization.text
            granularity: word or char
            stopwords: default or None or list

    errors:
        warn,
        raise,
        coerce,
        strict,
        ignore

    Returns
    -------
    list
        When tokenization is False, return SegmentList;
        When tokenization is True, return TokenList

    Examples
    --------
    >>> test_item = r"如图所示，则$\bigtriangleup ABC$的面积是$\SIFBlank$。$\FigureID{1}$"
    >>> tl = sif4sci(test_item)
    >>> tl
    ['如图所示', '\\bigtriangleup', 'ABC', '面积', '\\SIFBlank', \FigureID{1}]
    >>> tl.describe()
    {'t': 2, 'f': 2, 'g': 1, 'm': 1}
    >>> with tl.filter('fgm'):
    ...     tl
    ['如图所示', '面积']
    >>> with tl.filter(keep='t'):
    ...     tl
    ['如图所示', '面积']
    >>> with tl.filter():
    ...     tl
    ['如图所示', '\\bigtriangleup', 'ABC', '面积', '\\SIFBlank', \FigureID{1}]
    >>> tl.text_tokens
    ['如图所示', '面积']
    >>> tl.formula_tokens
    ['\\bigtriangleup', 'ABC']
    >>> tl.figure_tokens
    [\FigureID{1}]
    >>> tl.ques_mark_tokens
    ['\\SIFBlank']
    >>> sif4sci(test_item, symbol="gm", tokenization_params={"formula_params": {"method": "ast"}})
    ['如图所示', <Formula: \bigtriangleup ABC>, '面积', '[MARK]', '[FIGURE]']
    >>> sif4sci(test_item, symbol="tfgm")
    ['[TEXT]', '[FORMULA]', '[TEXT]', '[MARK]', '[TEXT]', '[FIGURE]']
    >>> sif4sci(test_item, symbol="gm",
    ... tokenization_params={"formula_params": {"method": "ast", "return_type": "list"}})
    ['如图所示', '\\bigtriangleup', 'A', 'B', 'C', '面积', '[MARK]', '[FIGURE]']
    >>> test_item_1 = {
    ...     "stem": r"若$x=2$, $y=\sqrt{x}$，则下列说法正确的是$\SIFChoice$",
    ...     "options": [r"$x < y$", r"$y = x$", r"$y < x$"]
    ... }
    >>> tls = [
    ...     sif4sci(e, symbol="gm",
    ...     tokenization_params={
    ...         "formula_params": {
    ...             "method": "ast", "return_type": "list", "ord2token": True, "var_numbering": True,
    ...             "link_variable": False}
    ...     })
    ...     for e in ([test_item_1["stem"]] + test_item_1["options"])
    ... ]
    >>> tls[1:]
    [['mathord_0', '<', 'mathord_1'], ['mathord_0', '=', 'mathord_1'], ['mathord_0', '<', 'mathord_1']]
    >>> link_formulas(*tls)
    >>> tls[1:]
    [['mathord_0', '<', 'mathord_1'], ['mathord_1', '=', 'mathord_0'], ['mathord_1', '<', 'mathord_0']]
    >>> from EduNLP.utils import dict2str4sif
    >>> test_item_1_str = dict2str4sif(test_item_1, tag_mode="head", add_list_no_tag=False)
    >>> test_item_1_str  # doctest: +ELLIPSIS
    '$\\SIFTag{stem}$...则下列说法正确的是$\\SIFChoice$$\\SIFTag{options}$$x < y$$\\SIFSep$$y = x$$\\SIFSep$$y < x$'
    >>> tl1 = sif4sci(test_item_1_str, symbol="gm",
    ... tokenization_params={"formula_params": {"method": "ast", "return_type": "list", "ord2token": True}})
    >>> tl1.get_segments()[0]
    ['\\SIFTag{stem}']
    >>> tl1.get_segments()[1:3]
    [['[TEXT_BEGIN]', '[TEXT_END]'], ['[FORMULA_BEGIN]', 'mathord', '=', 'textord', '[FORMULA_END]']]
    >>> tl1.get_segments(add_seg_type=False)[0:3]
    [['\\SIFTag{stem}'], ['mathord', '=', 'textord'], ['mathord', '=', 'mathord', '{ }', '\\sqrt']]
    >>> test_item_2 = {"options": [r"$x < y$", r"$y = x$", r"$y < x$"]}
    >>> test_item_2
    {'options': ['$x < y$', '$y = x$', '$y < x$']}
    >>> test_item_2_str = dict2str4sif(test_item_2, tag_mode="head", add_list_no_tag=False)
    >>> test_item_2_str
    '$\\SIFTag{options}$$x < y$$\\SIFSep$$y = x$$\\SIFSep$$y < x$'
    >>> tl2 = sif4sci(test_item_2_str, symbol="gms",
    ... tokenization_params={"formula_params": {"method": "ast", "return_type": "list"}})
    >>> tl2
    ['\\SIFTag{options}', 'x', '<', 'y', '[SEP]', 'y', '=', 'x', '[SEP]', 'y', '<', 'x']
    >>> tl2.get_segments(add_seg_type=False)
    [['\\SIFTag{options}'], ['x', '<', 'y'], ['[SEP]'], ['y', '=', 'x'], ['[SEP]'], ['y', '<', 'x']]
    >>> tl2.get_segments(add_seg_type=False, drop="s")
    [['\\SIFTag{options}'], ['x', '<', 'y'], ['y', '=', 'x'], ['y', '<', 'x']]
    >>> tl3 = sif4sci(test_item_1["stem"], symbol="gs")
    >>> tl3.text_segments
    [['说法', '正确']]
    >>> tl3.formula_segments
    [['x', '=', '2'], ['y', '=', '\\sqrt', '{', 'x', '}']]
    >>> tl3.figure_segments
    []
    >>> tl3.ques_mark_segments
    [['\\SIFChoice']]
    >>> test_item_3 = r"已知$y=x$，则以下说法中$\textf{正确,b}$的是"
    >>> tl4 = sif4sci(test_item_3)
    Warning: there is some chinese characters in formula!
    >>> tl4.text_segments
    [['已知'], ['说法', '中', '正确']]
    """
    try:
        if mode in [1, 2]:
            check_formula = True if mode == 1 else False
            sif, item_parser = is_sif(item, check_formula=check_formula, return_parser=True)
            if sif is not True:
                item = to_sif(item, parser=item_parser)
        elif mode != 0:
            raise KeyError(
                "Unknown mode %s, use only 0 or 1 or 2." % mode
            )

        ret = seg(item, figures, symbol)

        if tokenization is True:
            ret = tokenize(ret, **(tokenization_params if tokenization_params is not None else {}))

        return ret
    except Exception as e:  # pragma: no cover
        msg = traceback.format_exc()
        if errors == "warn":
            warnings.warn(msg)
        elif errors == "raise":
            raise e



# coding: utf-8
# 2021/8/1 @ tongshiwei

import pytest
from EduNLP.Tokenizer import get_tokenizer
from EduNLP.Pretrain import DisenQTokenizer


def test_tokenizer():
    with pytest.raises(KeyError):
        get_tokenizer("error")


def test_disenQTokenizer():
    tokenizer = DisenQTokenizer(max_length=10, tokenize_method="space")
    # with pytest.raises(RuntimeError):
    #     tokenizer("10 米 的 (2/5) = () 米 的 (1/2) .")

    test_items = [
        "10 米 的 (2/5) = () 米 的 (1/2) . 多 余 的 字",
        "-1 - 1",
        "5 % 2 + 3.14",
        "3.x",
        ".",
        "",
        "-1/2",
        "/",
        "1.2%",
    ]
    tokenizer.set_vocab(test_items)
    print(tokenizer.vocab_size)
    for item in test_items:
        token_item = tokenizer(item)
        print(token_item)

    test_item = tokenizer(test_items[0], padding=True)
    assert test_item["seq_idx"].shape[-1] == 10


def test_CharTokenizer():
    items = [{
        "stem": "文具店有 $600$ 本练习本，卖出一些后，还剩 $4$ 包，每包 $25$ 本，卖出多少本？",
        "options": ["1", "2"]
    }]
    tokenizer = get_tokenizer("char", stop_words=set("，？"))
    tokens = tokenizer(items, key=lambda x: x['stem'])
    ret = next(tokens)
    ans = ['文', '具', '店', '有', '$', '600', '$', '本', '练', '习', '本', '卖', '出', '一',
           '些', '后', '还', '剩', '$', '4', '$', '包', '每', '包', '$', '25', '$', '本', '卖', '出', '多', '少', '本']
    assert ret == ans


def test_SpaceTokenizer():
    items = ['文具店有 $600$ 本练习本，卖出一些后，还剩 $4$ 包，每包 $25$ 本，卖出多少本？']
    tokenizer = get_tokenizer("space", stop_words=[])
    tokens = tokenizer(items)
    ret = next(tokens)
    ans = ['文具店有', '$600$', '本练习本，卖出一些后，还剩', '$4$', '包，每包', '$25$', '本，卖出多少本？']
    assert ret == ans


def test_AstformulaTokenizer():
    items = ['文具店有 $600$ 本练习本，卖出一些后，还剩 $4$ 包，每包 $25$ 本，卖出多少本？']
    tokenizer = get_tokenizer("ast_formula")
    tokens = tokenizer(items)
    ret = next(tokens)
    ans = ['文具店', 'textord', 'textord', 'textord', '练习本', '卖出', '剩', 'textord', '包', '每包', 'textord', 'textord', '卖出']
    assert ret == ans


def test_PuretextTokenizer():
    items = ['文具店有 $600$ 本练习本，卖出一些后，还剩 $4$ 包，每包 $25$ 本，卖出多少本？']
    tokenizer = get_tokenizer("pure_text", stop_words=[])
    tokens = tokenizer(items)
    ret = next(tokens)
    ans = ['文具店', '600', '练习本', '卖出', '剩', '4', '包', '每包', '25', '卖出']
    assert ret == ans
    tokenizer = get_tokenizer("pure_text", stop_words=[], handle_figure_formula=None)
    tokens = tokenizer(items)
    ret = next(tokens)
    assert ret == ans
    tokenizer = get_tokenizer("pure_text", stop_words=[], handle_figure_formula='symbolize')
    tokens = tokenizer(items)
    ret = next(tokens)
    assert ret == ans
    with pytest.raises(ValueError):
        tokenizer = get_tokenizer("pure_text", stop_words=[], handle_figure_formula='wrong')


def test_CustomTokenizer():
    items = [{
        "stem": "文具店有 $600$ 本练习本，卖出一些后，还剩 $4$ 包，每包 $25$ 本，卖出多少本？",
        "options": ["1", "2"]
    }]
    tokenizer = get_tokenizer("custom", symbol='f')
    tokens = tokenizer(items, key=lambda x: x['stem'])
    ret = next(tokens)
    ans = ['文具店', '[FORMULA]', '练习本', '卖出', '剩', '[FORMULA]', '包', '每包', '[FORMULA]', '卖出']
    assert ret == ans
    items = [{
        "stem": "有公式$\\FormFigureID{1}$，如图$\\FigureID{088f15ea-xxx}$,若$x,y$满足约束条件公式$\\F\
            ormFigureBase64{2}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$",
        "options": ["1", "2"]
    }]
    tokenizer = get_tokenizer("custom", symbol='f', handle_figure_formula="symbolize")
    tokens = tokenizer(items, key=lambda x: x['stem'])
    ret = next(tokens)
    ret.pop(3)
    ans = ['公式', '[FORMULA]', '如图', '\\FigureID{088f15ea-xxx}', '[FORMULA]', '约束条件', '公式', '[FORMULA]',
           '\\SIFSep', '[FORMULA]', '最大值', '\\SIFBlank']
    ans.pop(3)
    assert ret == ans
